# a clone of the ruby example https://gist.github.com/valeriomazzeo/5491aee76f758f7352e2e6611ce87ec1
import os
from os import path

import re

import click
from click import Path
from github import Github, UnknownObjectException
# from valid8 import validate  not compliant with python 2.7


@click.command()
@click.option('-u', '--user', help='GitHub username')
@click.option('-p', '--pwd', help='GitHub password')
@click.option('-s', '--secret', help='GitHub access token')
@click.option('-r', '--repo-slug', help='Repo slug. i.e.: apple/swift')
@click.option('-cf', '--changelog-file', help='Changelog file path')
@click.option('-d', '--doc-url', help='Documentation url')
@click.option('-df', '--data-file', help='Data file to upload', type=Path(exists=True, file_okay=True, dir_okay=False,
                                                                          resolve_path=True))
@click.argument('tag')
def create_or_update_release(user, pwd, secret, repo_slug, changelog_file, doc_url, data_file, tag):
    """
    Creates or updates (TODO)
    a github release corresponding to git tag <TAG>.
    """
    # 1- AUTHENTICATION
    if user is not None and secret is None:
        # using username and password
        # validate('user', user, instance_of=str)
        assert isinstance(user, str)
        # validate('pwd', pwd, instance_of=str)
        assert isinstance(pwd, str)
        g = Github(user, pwd)
    elif user is None and secret is not None:
        # or using an access token
        # validate('secret', secret, instance_of=str)
        assert isinstance(secret, str)
        g = Github(secret)
    else:
        raise ValueError("You should either provide username/password OR an access token")
    click.echo("Logged in as {user_name}".format(user_name=g.get_user()))

    # 2- CHANGELOG VALIDATION
    regex_pattern = "[\s\S]*[\n][#]+[\s]*(?P<title>[\S ]*%s[\S ]*)[\n]+?(?P<body>[\s\S]*?)[\n]*?(\n#|$)" % re.escape(tag)
    changelog_section = re.compile(regex_pattern)
    if changelog_file is not None:
        # validate('changelog_file', changelog_file, custom=os.path.exists,
        #          help_msg="changelog file should be a valid file path")
        assert os.path.exists(changelog_file), "changelog file should be a valid file path"
        with open(changelog_file) as f:
            contents = f.read()

        match = changelog_section.match(contents).groupdict()
        if match is None or len(match) != 2:
            raise ValueError("Unable to find changelog section matching regexp pattern in changelog file.")
        else:
            title = match['title']
            message = match['body']
    else:
        title = tag
        message = ''

    # append footer if doc url is provided
    message += "\n\nSee [documentation page](%s) for details." % doc_url

    # 3- REPOSITORY EXPLORATION
    # validate('repo_slug', repo_slug, instance_of=str, min_len=1, help_msg="repo_slug should be a non-empty string")
    assert isinstance(repo_slug, str) and len(repo_slug) > 0, "repo_slug should be a non-empty string"
    repo = g.get_repo(repo_slug)

    # -- Is there a tag with that name ?
    try:
        tag_ref = repo.get_git_ref("tags/" + tag)
    except UnknownObjectException:
        raise ValueError("No tag with name %s exists in repository %s" % (tag, repo.name))

    # -- Is there already a release with that tag name ?
    click.echo("Checking if release %s already exists in repository %s" % (tag, repo.name))
    try:
        release = repo.get_release(tag)
        if release is not None:
            raise ValueError("Release %s already exists in repository %s. Please set overwrite to True if you wish to "
                             "update the release (Not yet supported)" % (tag, repo.name))
    except UnknownObjectException:
        # Release does not exist: we can safely create it.
        click.echo("Creating release %s on repo: %s" % (tag, repo.name))
        click.echo("Release title: '%s'" % title)
        click.echo("Release message:\n--\n%s\n--\n" % message)
        repo.create_git_release(tag=tag, name=title,
                                message=message,
                                draft=False, prerelease=False)

        # add the asset file if needed
        if data_file is not None:
            release = None
            while release is None:
                release = repo.get_release(tag)
            release.upload_asset(path=data_file, label=path.split(data_file)[1], content_type="application/gzip")

        # --- Memo ---
        # release.target_commitish  # 'master'
        # release.tag_name    # '0.5.0'
        # release.title       # 'First public release'
        # release.body        # markdown body
        # release.draft       # False
        # release.prerelease  # False
        # #
        # release.author
        # release.created_at  # datetime.datetime(2018, 11, 9, 17, 49, 56)
        # release.published_at  # datetime.datetime(2018, 11, 9, 20, 11, 10)
        # release.last_modified  # None
        # #
        # release.id           # 13928525
        # release.etag         # 'W/"dfab7a13086d1b44fe290d5d04125124"'
        # release.url          # 'https://api.github.com/repos/smarie/python-odsclient/releases/13928525'
        # release.html_url     # 'https://github.com/smarie/python-odsclient/releases/tag/0.5.0'
        # release.tarball_url  # 'https://api.github.com/repos/smarie/python-odsclient/tarball/0.5.0'
        # release.zipball_url  # 'https://api.github.com/repos/smarie/python-odsclient/zipball/0.5.0'
        # release.upload_url   # 'https://uploads.github.com/repos/smarie/python-odsclient/releases/13928525/assets{?name,label}'


if __name__ == '__main__':
    create_or_update_release()



import os

import pytest
import requests

from odsclient import get_whole_dataset, ODSException, NoODSAPIKeyFoundError, InsufficientRightsForODSResourceError
from odsclient.core import baseurl_to_id_str


def test_error_bad_dataset_id():
    """Tests that an error associated with bad dataset id is correctly parsed and raised as an ODSException"""
    with pytest.raises(ODSException) as exc_info:
        get_whole_dataset("unknwn", platform_id='public')

    # see https://github.com/psf/requests/blob/master/requests/status_codes.py
    assert exc_info.value.status_code == requests.codes.NOT_FOUND  # not found
    assert exc_info.value.error_msg == "Unknown dataset: unknwn"


def test_no_apikey_provided():
    """Tests that enforce_apikey works correctly"""
    with pytest.raises(NoODSAPIKeyFoundError):
        get_whole_dataset("world-growth-since-the-industrial-revolution0", enforce_apikey=True)


def test_apikey_not_granting_rights():
    """Tests that if rights are not sufficient the proper error is raised"""
    with pytest.raises(InsufficientRightsForODSResourceError):
        get_whole_dataset("employment-by-sector-in-france-and-the-united-states-1800-2012",
                          base_url="https://data.exchange.se.com/")


def test_bad_apikey():
    """Tests that an error associated with bad api key is correctly parsed and raised as an ODSException"""
    with pytest.raises(ODSException) as exc_info:
        get_whole_dataset("world-growth-since-the-industrial-revolution0",
                          apikey="my_non_working_api_key")

    assert exc_info.value.status_code == requests.codes.UNAUTHORIZED  # not authorized
    assert exc_info.value.error_msg == "API key is not valid"


# @pytest.mark.skipif('TRAVIS_PYTHON_VERSION' in os.environ, reason="Does not work yet on travis")
def test_keyring_unit():
    """Small unit test for keyring"""
    import keyring
    print(keyring.get_keyring())
    base_url = "https://data.exchange.se.com/"
    keyring.set_password(base_url, 'apikey', 'blah')
    assert keyring.get_password(base_url, 'apikey') == 'blah'


@pytest.mark.parametrize("protocol", ["http://", "ftp://", "https://"])
@pytest.mark.parametrize("ending_slash", [False, True])
def test_baseurl_to_id_str(protocol, ending_slash):
    base_url = protocol + "public.opendatasoft.com"
    if ending_slash:
        base_url += "/"
    pseudo_id = baseurl_to_id_str(base_url)
    assert pseudo_id == "public"

    base_url = protocol + "data.exchange.se.com/ho"
    if ending_slash:
        base_url += "/"
    pseudo_id = baseurl_to_id_str(base_url)
    assert pseudo_id == "data.exchange.se.com_ho"



from subprocess import call

try:
    import click
except ImportError as e:
    raise Exception("`click` is required for `odskeys` to work. Please `pip install click`. Caught: %r" % e)

from odsclient.core import KR_DEFAULT_USERNAME, ODSClient
from odsclient.shortcuts import store_apikey_in_keyring, get_apikey_from_keyring, remove_apikey_from_keyring


@click.group()
def odskeys():
    """
    Commandline utility to get/set/remove api keys from the OS keyring using the `keyring` library.
    To get help on each command use:

        odskeys <cmd> --help

    """
    pass


def _get_url_used(platform_id,  # type: str
                  base_url,     # type: str
                  ):
    # type: (...) -> str
    """ returns the url actually used by the ODSClient """
    client = ODSClient(platform_id=platform_id, base_url=base_url)
    return client.base_url


@odskeys.command(name="get")
@click.option('-p', '--platform_id', default='public', help="Specific ODS platform id. Default 'public'")
@click.option('-b', '--base_url', default=None, help="Specific ODS base url. Default: "
                                                     "https://<platform_id>.opendatasoft.com/")
@click.option('-u', '--username', default=KR_DEFAULT_USERNAME, help='Custom username to use in the keyring entry. '
                                                                    'Default: %s' % KR_DEFAULT_USERNAME)
def get_ods_apikey(platform_id,                   # type: str
                   base_url,                      # type: str
                   username=KR_DEFAULT_USERNAME,  # type: str
                   ):
    """
    Looks up an ODS apikey entry in the keyring. Custom ODS platform id or base url can be provided through options.
    """
    apikey = get_apikey_from_keyring(platform_id=platform_id,
                                     base_url=base_url,
                                     keyring_entries_username=username,
                                     )
    # actual url used, for message prints
    url_used = _get_url_used(platform_id=platform_id, base_url=base_url)
    if apikey is not None:
        click.echo("Api key found for platform url '%s': %s" % (url_used, apikey))
    else:
        click.echo("No api key registered for platform url '%s'" % (url_used, ))


@odskeys.command(name="remove")
@click.option('-p', '--platform_id', default='public', help="Specific ODS platform id. Default 'public'")
@click.option('-b', '--base_url', default=None, help="Specific ODS base url. Default: "
                                                     "https://<platform_id>.opendatasoft.com/")
@click.option('-u', '--username', default=KR_DEFAULT_USERNAME, help='Custom username to use in the keyring entry. '
                                                                    'Default: %s' % KR_DEFAULT_USERNAME)
def remove_ods_apikey(platform_id='public',          # type: str
                      base_url=None,                 # type: str
                      username=KR_DEFAULT_USERNAME,  # type: str
                      ):
    """
    Removes an ODS apikey entry from the keyring. Custom ODS platform id or base url can be provided through options.
    """
    apikey = get_apikey_from_keyring(platform_id=platform_id,
                                     base_url=base_url,
                                     keyring_entries_username=username,
                                     )
    # actual url used, for message prints
    url_used = _get_url_used(platform_id=platform_id, base_url=base_url)
    if apikey is None:
        click.echo("No api key registered for platform url '%s'" % (url_used,))
    else:
        remove_apikey_from_keyring(platform_id=platform_id,
                                   base_url=base_url,
                                   keyring_entries_username=username,
                                   )
        apikey = get_apikey_from_keyring(platform_id=platform_id,
                                         base_url=base_url,
                                         keyring_entries_username=username,
                                         )
        assert apikey is None
        click.echo("Api key removed successfully for platform url '%s'" % (url_used,))


@odskeys.command(name="set")
@click.option('-p', '--platform_id', default='public', help="Specific ODS platform id. Default 'public'")
@click.option('-b', '--base_url', default=None, help="Specific ODS base url. Default: "
                                                     "https://<platform_id>.opendatasoft.com/")
@click.option('-u', '--username', default=KR_DEFAULT_USERNAME, help='Custom username to use in the keyring entry. '
                                                                    'Default: %s' % KR_DEFAULT_USERNAME)
@click.option('-k', '--apikey', help="apikey to register. If none is provided, you will be prompted for it.")
def set_ods_apikey(platform_id='public',          # type: str
                   base_url=None,                 # type: str
                   username=KR_DEFAULT_USERNAME,  # type: str
                   apikey=None                    # type: str
                   ):
    """
    Creates an ODS apikey entry in the keyring. Custom ODS platform id or base url can be provided through options.
    """
    store_apikey_in_keyring(platform_id=platform_id,
                            base_url=base_url,
                            keyring_entries_username=username,
                            apikey=apikey
                            )
    gotapikey = get_apikey_from_keyring(platform_id=platform_id,
                                        base_url=base_url,
                                        keyring_entries_username=username,
                                        )
    if apikey is not None:
        assert apikey == gotapikey
    else:
        # api key provided through getpass() - we do not have access to it
        assert gotapikey is not None
    # actual url used, for message print
    url_used = _get_url_used(platform_id=platform_id, base_url=base_url)
    click.echo("Api key defined successfully for platform url '%s'" % (url_used,))


@odskeys.command(name="show")
@click.option('-a', '--alt', default=0, help='Alternative #id (the number of available alternatives is '
                                             'platform dependent)')
def show_os_mgr(alt=0,  # type: int
                ):
    """
    Shows the OS credentials manager associated with current keyring backend.
    When several alternatives exist, the `--alt` option can be used to switch from 0 (default) to others.

    Currently supported backends and alternatives:

    Windows WinVaultKeyring:
     - (alt 0, default) control.exe /name Microsoft.CredentialManager
     - (alt 1) rundll32.exe keymgr.dll, KRShowKeyMgr

    """
    import keyring
    kr = keyring.get_keyring()
    if 'Windows WinVaultKeyring' in kr.name:
        alts = {
            0: ["control.exe", "/name", "Microsoft.CredentialManager"],
            1: ["cmd.exe", "/c", "start", "/B", "rundll32.exe", "keymgr.dll,KRShowKeyMgr"]
        }
    else:
        click.echo("This command is not supported for keyring backend '%s', please report it here:"
                   " https://github.com/smarie/python-odsclient/issues/" % (kr.name, ))
        return

    # execute the alternative
    try:
        cmd = alts[alt]
    except KeyError:
        click.echo("Invalid alternative #: %s. Only [0-%s] are supported with keyring backend %s"
                   % (alt, len(alts)-1, kr.name))
        return
    else:
        click.echo("Keyring backend is '%s'. Runnning command for alternative %s: '%s'" % (kr.name, alt, ' '.join(cmd)))
        call(cmd)


# @odskeys.command(name="list")
# @click.option('-p', '--platform_id', default='public', help='Filter on a specific ODS platform id.')
# @click.option('-b', '--base_url', default=1, help='Filter on a specific ODS base url')
# def list_ods_keys(platform_id,  # type: str
#                   base_url      # type: str
#                   ):
#     """
#     Lists the
#     """
#     click.echo('Filtering: %s, %s!' % (platform_id, base_url))
#     # list with https://github.com/jaraco/keyring/issues/151
#     # but this does not work with windows backend
#     for item in keyring.get_keyring().get_preferred_collection().get_all_items():
#         print(item.get_label(), item.get_attributes())


if __name__ == '__main__':
    odskeys()



# coding: utf-8
import pandas as pd

from odsclient.utils import create_reading_buffer


def ref_dataset_public_platform():
    """Return a reference dataset for the public ODS platform """

    dataset_id = "respect-des-delais-dacheminement-courrier"

    ref_csv = """Catégorie;Objectif ou Réalisation;Pourcentage;Annee
Lettre prioritaire J + 1;réalisation;0.867;2014
Colissimo guichet J + 2;objectif;0.88;2014
Courrier industriel J + 2;réalisation;0.965;2009
Lettre prioritaire J + 1;réalisation;0.847;2009
Lettre prioritaire J + 1;objectif;0.85;2014
Lettre recommandée J + 2;objectif;0.9400000000000001;2014
Lettre recommandée J + 2;réalisation;0.887;2009
Courrier industriel J + 2;réalisation;0.969;2014
Colissimo guichet J + 2;réalisation;0.91;2014
Courrier industriel J + 2;objectif;0.9500000000000001;2009
Colissimo guichet J + 2;réalisation;0.877;2009
Lettre verte J + 2;objectif;0.9400000000000001;2014
Lettre prioritaire J + 1;objectif;0.84;2009
Lettre recommandée J + 2;réalisation;0.9460000000000001;2014
Lettre verte J + 2;réalisation;0.932;2014
Courrier industriel J + 2;objectif;0.9500000000000001;2014
Colissimo guichet J + 2;objectif;0.86;2009
""".replace("\n", "\r\n")

    # we keep this explicit just in case the ref_df reading does not work as expected
    ref_shape = (17, 1)

    ref_df = pd.read_csv(create_reading_buffer(ref_csv, is_literal=True), sep=';')
    ref_df = ref_df.set_index(['Catégorie', 'Objectif ou Réalisation', 'Annee']).sort_index()

    return dataset_id, ref_csv, ref_df, ref_shape


def ref_dataset_other_platform():
    """ Return a reference dataset for the uat-data.exchange.se.com ODS platform """

    # shared info
    # dataset_id = "employment-by-sector-in-france-and-the-united-states-1800-2012"
    # base_url = "https://data.exchange.se.com/"
    dataset_id = "odsclient-reference-dataset"
    base_url = "https://uat-data.exchange.se.com/"

    ref_csv = """Transport;Année;Millions de Voyageurs
SNCF - Trains/RER (y compris T4);;
RATP - RER;;
RATP - RER;2011;7575
RATP - Métro;;
RATP - RER;2014;7722
RATP - Métro;2014;5194
SNCF - Trains/RER (y compris T4);2011;11583
RATP - RER;2010;7486
RATP - RER;2013;7605
SNCF - Trains/RER (y compris T4);2013;12103
RATP - Métro;2012;5130
RATP - RER;2012;7675
SNCF - Trains/RER (y compris T4);2014;12148
SNCF - Trains/RER (y compris T4);2010;11221
RATP - Métro;2013;5044
RATP - Métro;2010;4892
SNCF - Trains/RER (y compris T4);2012;11816
RATP - Métro;2011;5022
""".replace("\n", "\r\n")
    ref_df = pd.read_csv(create_reading_buffer(ref_csv, is_literal=True), sep=';')
    ref_df = ref_df.set_index(['Transport', 'Année']).sort_index()

    ref_shape = (18, 1)

    return base_url, dataset_id, ref_csv, ref_df, ref_shape



import sys

from io import BytesIO   # to handle byte strings
from io import StringIO  # to handle unicode strings

if sys.version_info >= (3, 0):
    def create_reading_buffer(value, is_literal):
        return StringIO(value)
else:
    def create_reading_buffer(value, is_literal):
        if is_literal:
            return BytesIO(value)
        else:
            return StringIO(value)



from odsclient.core import ODSClient, ODSException, NoODSAPIKeyFoundError, InsufficientRightsForODSResourceError, \
    ENV_ODS_APIKEY, KR_DEFAULT_USERNAME, CacheEntry
from odsclient.shortcuts import get_whole_dataset, get_whole_dataframe, store_apikey_in_keyring, \
    get_apikey_from_keyring, remove_apikey_from_keyring, get_apikey, clean_cache, get_cached_dataset_entry

__all__ = [
    # submodules
    'core', 'shortcuts',
    # symbols
    'ODSClient', 'ODSException', 'NoODSAPIKeyFoundError', 'InsufficientRightsForODSResourceError',
    'ENV_ODS_APIKEY', 'KR_DEFAULT_USERNAME',
    'get_whole_dataset', 'get_whole_dataframe', 'store_apikey_in_keyring', 'get_apikey_from_keyring',
    'remove_apikey_from_keyring', 'get_apikey', 'clean_cache', 'get_cached_dataset_entry', 'CacheEntry'
]



# coding: utf-8
from __future__ import print_function

import os
import sys

import pytest
import pandas as pd
import keyring
from requests import Session
from requests.exceptions import ProxyError

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2

if sys.version_info < (3,):
    # We need a version of open that supports encoding and nline endings
    # See https://stackoverflow.com/a/10975371/7262247
    from io import open

from .ref_datasets import ref_dataset_public_platform, create_reading_buffer, ref_dataset_other_platform

from odsclient import get_whole_dataset, get_whole_dataframe, store_apikey_in_keyring, \
    remove_apikey_from_keyring, get_apikey, clean_cache, get_cached_dataset_entry


DEFAULT_CACHE_ROOT = ".odsclient"
clean_cache()

ALT_CACHE_ROOT = ".odscustcache"
clean_cache(cache_root=ALT_CACHE_ROOT)


def test_invalid_network_connection():
    """Tests that the make_invalid_network_session helper function works as expected"""
    with pytest.raises(ProxyError):
        get_whole_dataset(dataset_id="fake", requests_session=make_invalid_network_session())


@pytest.mark.parametrize("save_to_file", [False, True], ids="save_to_file={}".format)
@pytest.mark.parametrize("progress_bar", [False, True], ids="progress_bar={}".format)
@pytest.mark.parametrize("file_cache", [False, True, ALT_CACHE_ROOT], ids="file_cache={}".format)
def test_example(save_to_file, progress_bar, tmp_path, file_cache):
    """basic test: retrieve an example dataset """

    # get the reference dataset
    dataset_id, ref_csv, ref_df, ref_shape = ref_dataset_public_platform()

    # check the cache status
    cached_entry = None
    if file_cache:
        cache_root = DEFAULT_CACHE_ROOT if file_cache is True else file_cache
        cached_entry = get_cached_dataset_entry(dataset_id=dataset_id, platform_id="public",
                                                cache_root=None if file_cache is True else cache_root)
        assert cached_entry.file_path.as_posix() == "%s/public/%s.csv" % (cache_root, dataset_id)
        # the cache is supposed to be clean (cleaned at the end of this test)
        assert not cached_entry.exists()

    # with debug_requests():
    to_path = tmp_path / "blah" / "tmp.csv" if save_to_file else None
    csv_str = get_whole_dataset(dataset_id, platform_id='public', file_cache=file_cache, to_path=to_path,
                                tqdm=progress_bar)

    if save_to_file:
        assert csv_str is None
        # note: newline='' preserves line ending while opening. See https://stackoverflow.com/a/50996542/7262247
        with open(str(to_path), mode="rt", encoding="utf-8", newline='') as f:
            csv_str = f.read()
        os.remove(str(to_path))

    # compare the text string (if order does not change across queries...)
    # assert csv_str == ref_csv

    # move to pandas
    df = pd.read_csv(create_reading_buffer(csv_str, is_literal=False), sep=';')

    # compare with ref
    df = df.set_index(['Catégorie', 'Objectif ou Réalisation', 'Annee']).sort_index()
    pd.testing.assert_frame_equal(df, ref_df)
    assert df.shape == ref_shape

    # test the pandas direct streaming API without cache
    df2 = get_whole_dataframe(dataset_id, tqdm=progress_bar)
    df2 = df2.set_index(['Catégorie', 'Objectif ou Réalisation', 'Annee']).sort_index()
    pd.testing.assert_frame_equal(df, df2)

    # make sure the cached entry exists now and can be read without internet connection
    if cached_entry:
        # Make sure that the cache entry contains the dataset
        assert cached_entry.exists()
        # note: newline='' preserves line ending while opening. See https://stackoverflow.com/a/50996542/7262247
        with open(str(cached_entry.file_path), mode="rt", encoding="utf-8", newline='') as f:
            cached_csv_str = f.read()
        assert cached_csv_str == csv_str

        # New offline query: the cache should be hit even if the public platform is identified using its base_url here
        csv_str2 = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache,
                                     base_url="https://public.opendatasoft.com", tqdm=progress_bar,
                                     requests_session=make_invalid_network_session())
        assert csv_str2 == cached_csv_str

        # Same with the other method
        df3 = get_whole_dataframe(dataset_id, file_cache=file_cache, requests_session=make_invalid_network_session(),
                                  tqdm=progress_bar)
        df3 = df3.set_index(['Catégorie', 'Objectif ou Réalisation', 'Annee']).sort_index()
        pd.testing.assert_frame_equal(df, df3)

        # clean it for next time
        cached_entry.delete()
        assert not cached_entry.exists()

        # Make sure it is re-cached if we use the dataframe-getter method directly
        df4 = get_whole_dataframe(dataset_id, file_cache=file_cache, tqdm=progress_bar)
        df4 = df4.set_index(['Catégorie', 'Objectif ou Réalisation', 'Annee']).sort_index()
        assert cached_entry.exists()
        pd.testing.assert_frame_equal(df, df4)

        # clean it for next time
        cached_entry.delete()
        assert not cached_entry.exists()


apikey_methods = ['direct', 'file_default', 'file_custom',
                  # 'multi_env_pfid', not available on this ODS target
                  'single_env', 'multi_env_baseurl', 'multi_env_default',
                  'keyring1', 'keyring2']


@pytest.mark.parametrize("apikey_method", apikey_methods)
@pytest.mark.parametrize("file_cache", [False, True, ALT_CACHE_ROOT], ids="file_cache={}".format)
def test_other_platform(apikey_method, file_cache):
    """Tests that the lib can connect to a different ODS platform with api key and custom url"""

    # get the reference dataset
    base_url, dataset_id, ref_csv, ref_df, ref_shape = ref_dataset_other_platform()

    test_apikey = os.environ['EXCH_AKEY']  # <-- travis

    # check the cache status
    cached_entry = None
    if file_cache:
        cache_root = DEFAULT_CACHE_ROOT if file_cache is True else file_cache
        cached_entry = get_cached_dataset_entry(dataset_id=dataset_id, base_url=base_url,
                                                cache_root=None if file_cache is True else cache_root)
        assert cached_entry.file_path.as_posix() == "%s/uat-data.exchange.se.com/%s.csv" % (cache_root, dataset_id)
        # the cache is supposed to be clean (cleaned at the end of this test)
        assert not cached_entry.exists()

    # various methods to get the api key
    if apikey_method == 'direct':
        csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url, apikey=test_apikey)

    elif apikey_method == 'file_default':
        f_name = 'ods.apikey'
        assert not os.path.exists(f_name), "File '%s' already exists, please delete it first" % f_name
        with open(f_name, 'wb') as f:
            f.write(test_apikey.encode("utf-8"))

        assert get_apikey(base_url=base_url) == test_apikey
        try:
            csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url)
        finally:
            os.remove(f_name)

    elif apikey_method == 'file_custom':
        f_name = 'tmp.tmp'
        assert not os.path.exists(f_name), "File '%s' already exists, please delete it first" % f_name
        with open(f_name, 'wb') as f:
            f.write(test_apikey.encode("utf-8"))
        assert get_apikey(base_url=base_url, apikey_filepath=f_name) == test_apikey
        try:
            csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url,
                                        apikey_filepath=f_name)
        finally:
            os.remove(f_name)

    elif apikey_method == 'single_env':
        os.environ['ODS_APIKEY'] = test_apikey
        assert get_apikey(base_url=base_url) == test_apikey
        csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url)
        del os.environ['ODS_APIKEY']

    elif apikey_method == 'multi_env_baseurl':
        os.environ['ODS_APIKEY'] = "{'default': 'blah', '%s': '%s'}" % (base_url, test_apikey)
        assert get_apikey(base_url=base_url) == test_apikey
        csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url)
        del os.environ['ODS_APIKEY']

    elif apikey_method == 'multi_env_default':
        os.environ['ODS_APIKEY'] = "{'default': '%s', 'other_id': 'blah'}" % (test_apikey)
        assert get_apikey(base_url=base_url) == test_apikey
        csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url)
        del os.environ['ODS_APIKEY']

    elif apikey_method == 'keyring1':
        # if 'TRAVIS_PYTHON_VERSION' in os.environ:
        #     pytest.skip("Does not work yet on travis")
        keyring.set_password(base_url, 'apikey', test_apikey)
        assert get_apikey(base_url=base_url, keyring_entries_username='apikey') == test_apikey
        csv_str = get_whole_dataset(dataset_id=dataset_id, base_url=base_url, file_cache=file_cache,
                                    keyring_entries_username='apikey')
        keyring.delete_password(base_url, 'apikey')
        assert keyring.get_password(base_url, 'apikey') is None

    elif apikey_method == 'keyring2':
        # if 'TRAVIS_PYTHON_VERSION' in os.environ:
        #     pytest.skip("Does not work yet on travis")
        store_apikey_in_keyring(base_url=base_url, keyring_entries_username='apikey', apikey=test_apikey)
        assert get_apikey(base_url=base_url, keyring_entries_username='apikey') == test_apikey
        csv_str = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url,
                                    keyring_entries_username='apikey')
        remove_apikey_from_keyring(base_url=base_url, keyring_entries_username='apikey')
        assert keyring.get_password(base_url, 'apikey') is None
        
    else:
        raise ValueError('wrong apikey_method: %s' % apikey_method)

    # do not compare csv_str to ref_csv as order may change
    # assert csv_str == ref_csv

    # move to pandas
    df = pd.read_csv(create_reading_buffer(csv_str, is_literal=False), sep=';')

    # compare with ref
    df = df.set_index(['Transport', 'Année']).sort_index()
    pd.testing.assert_frame_equal(df, ref_df)
    assert df.shape == ref_shape

    # make sure the cached entry exists now and can be read without network connection
    if cached_entry:
        # make sure that the cache entry contains the dataset
        assert cached_entry.read() == csv_str

        # perform a second query offline and make sure the cache is hit
        csv_str2 = get_whole_dataset(dataset_id=dataset_id, file_cache=file_cache, base_url=base_url,
                                     requests_session=make_invalid_network_session())
        assert csv_str2 == csv_str

        # clean it for the next time
        # cached_entry.delete()
        clean_cache(dataset_id=dataset_id, base_url=base_url, cache_root=None if file_cache is True else cache_root)


def make_invalid_network_session():
    """Returns a session with an invalid proxy"""
    offline_session = Session()
    offline_session.proxies = {
        "http": "http://localhost:44445",
        "https": "http://localhost:44445"
    }
    offline_session.trust_env = False
    return offline_session



import sys
try:
    # python 3
    from urllib.parse import quote_plus
except ImportError:
    # python 2
    from urllib import quote_plus

import requests
import shutil
from os import makedirs, path
import xunitparser


class TestStats(object):
    def __init__(self, success_percentage, success, runned, skipped):
        self.success_percentage = success_percentage
        self.success = success
        self.runned = runned
        self.skipped = skipped


def get_test_stats(junit_xml='reports/junit/junit.xml'  # type: str
                   ):
    # type: (...) -> TestStats
    """
    read the junit test file and extract the success percentage
    :param junit_xml: the junit xml file path
    :return: the success percentage (an int)
    """
    ts, tr = xunitparser.parse(open(junit_xml))
    skipped = len(tr.skipped)
    runned = tr.testsRun - skipped
    failed = len(tr.failures)
    success = runned - failed

    success_percentage = round(success * 100 / runned)

    return TestStats(success_percentage, success, runned, skipped)


def download_badge(test_stats,                   # type: TestStats
                   dest_folder='reports/junit',  # type: str
                   badge_name='junit-badge.svg'  # type: str
                   ):
    """
    Downloads the badge corresponding to the provided success percentage, from https://img.shields.io.

    :param test_stats:
    :param dest_folder:
    :return:
    """
    if not path.exists(dest_folder):
        makedirs(dest_folder)  # , exist_ok=True) not python 2 compliant

    if test_stats.success_percentage < 50:
        color = 'red'
    elif test_stats.success_percentage < 75:
        color = 'orange'
    elif test_stats.success_percentage < 90:
        color = 'green'
    else:
        color = 'brightgreen'

    left_txt = "tests"
    # right_txt = "%s%%" % test_stats.success_percentage
    right_txt = "%s/%s" % (test_stats.success, test_stats.runned)
    url = 'https://img.shields.io/badge/%s-%s-%s.svg' % (left_txt, quote_plus(right_txt), color)

    dest_file = path.join(dest_folder, badge_name)

    print('Generating junit badge from : ' + url)
    response = requests.get(url, stream=True)
    with open(dest_file, 'wb') as out_file:
        response.raw.decode_content = True
        shutil.copyfileobj(response.raw, out_file)
    del response


if __name__ == "__main__":
    # Execute only if run as a script.
    # Check the arguments
    nbargs = len(sys.argv[1:])
    if nbargs < 1:
        raise ValueError("a mandatory argument is required: <threshold>, and an optional: <dest_folder>")
    else:
        threshold = float(sys.argv[1])
        if nbargs < 2:
            dest_folder = None
        elif nbargs < 3:
            dest_folder = sys.argv[2]
        else:
            raise ValueError("too many arguments received: 2 maximum (<threshold>, <dest_folder>)")

    # First retrieve the success percentage from the junit xml
    test_stats = get_test_stats(junit_xml='%s/junit.xml' % ('reports/junit' if dest_folder is None else dest_folder))

    # Validate against the threshold
    print("Success percentage is %s%%. Checking that it is >= %s" % (test_stats.success_percentage, threshold))
    if test_stats.success_percentage < threshold:
        raise Exception("Success percentage %s%% is strictly lower than required threshold %s%%"
                        "" % (test_stats.success_percentage, threshold))

    # Download the badge
    download_badge(test_stats, dest_folder=dest_folder)



import asyncio
from collections import namedtuple
from inspect import signature, isfunction
import logging
from pathlib import Path
import shutil
import subprocess
import sys
import os

from typing import Sequence, Dict, Union, Iterable, IO, Mapping, Tuple

from makefun import wraps, remove_signature_parameters

import nox
from nox.sessions import Session


nox_logger = logging.getLogger("nox")


PY27, PY35, PY36, PY37, PY38 = "2.7", "3.5", "3.6", "3.7", "3.8"
DONT_INSTALL = "dont_install"


def power_session(
        func=None,
        envs=None,
        python=None,
        py=None,
        reuse_venv=None,
        name=None,
        venv_backend=None,
        venv_params=None,
        logsdir=None,
        **kwargs
):
    """A nox.session on steroids"""
    if func is not None:
        return power_session()(func)
    else:
        if envs is not None:
            if py is not None or python is not None:
                raise ValueError("Only one of `envs` and `py/python` should be provided")
            python = list(envs.keys())
        nox_deco = nox.session(python=python, py=py, reuse_venv=reuse_venv, name=name, venv_backend=venv_backend,
                               venv_params=venv_params, **kwargs)

        def combined_decorator(f):
            # apply all decorators in turn
            f = with_power_session(f)

            # @with_logfile
            if logsdir is not None:
                f = with_logfile(logs_dir=logsdir)(f)

            # @inject_envs_params
            if envs is not None:
                f = inject_envs_params(envs)(f)

            # finally @nox.session
            return nox_deco(f)

        return combined_decorator


def with_power_session(f=None):
    """ A decorator to patch the session objects in order to add all methods from Session2"""

    if f is not None:
        return with_power_session()(f)

    def _decorator(f):
        @wraps(f)
        def _f_wrapper(**kwargs):
            # patch the session arg
            PowerSession.patch(kwargs['session'])

            # finally execute the session
            return f(**kwargs)

        return _f_wrapper

    return _decorator


class PowerSession(Session):
    """
    Our nox session improvements
    """

    # ------------ commandline runners -----------

    def run2(self,
             command: Union[Iterable[str], str],
             logfile: Union[bool, str, Path] = True,
             **kwargs):
        """
        An improvement of session.run that is able to

         - automatically split the provided command if it is a string
         - use a log file

        :param command:
        :param logfile: None/False (normal nox behaviour), or True (using nox file handler), or a file path.
        :param kwargs:
        :return:
        """
        if isinstance(command, str):
            command = command.split(' ')

        self.run(*command, logfile=logfile, **kwargs)

    def run_multi(self,
                  cmds: str,
                  logfile: Union[bool, str, Path] = True,
                  **kwargs):
        """
        An improvement of session.run that is able to

         - support multiline strings
         - use a log file

        :param cmds:
        :param logfile: None/False (normal nox behaviour), or True (using nox file handler), or a file path.
        :param kwargs:
        :return:
        """
        for cmdline in (line for line in cmds.splitlines() if line):
            self.run2(cmdline, logfile=logfile, **kwargs)

    # ------------ requirements installers -----------

    def install_reqs(
            self,
            # pre wired phases
            setup=False,
            install=False,
            tests=False,
            # custom phase
            phase=None,
            phase_reqs=None,
            versions_dct=None
    ):
        """
        A high-level helper to install requirements from the various project files

         - pyproject.toml "[build-system] requires" (if setup=True)
         - setup.cfg "[options] setup_requires" (if setup=True)
         - setup.cfg "[options] install_requires" (if install=True)
         - setup.cfg "[options] test_requires" (if tests=True)

        Two additional mechanisms are provided in order to customize how packages are installed.

        Conda packages
        --------------
        If the session runs on a conda environment, you can add a [tool.conda] section to your pyproject.toml. This
        section should contain a `conda_packages` entry containing the list of package names that should be installed
        using conda instead of pip.

        ```
        [tool.conda]
        # Declare that the following packages should be installed with conda instead of pip
        # Note: this includes packages declared everywhere, here and in setup.cfg
        conda_packages = [
            "setuptools",
            "wheel",
            "pip"
        ]
        ```

        Version constraints
        -------------------
        In addition to the version constraints in the pyproject.toml and setup.cfg, you can specify additional temporary
        constraints with the `versions_dct` argument , for example if you know that this executes on a specific python
        version that requires special care.
        For this, simply pass a dictionary of {'pkg_name': 'pkg_constraint'} for example {"pip": ">10"}.

        """

        # Read requirements from pyproject.toml
        toml_setup_reqs, toml_use_conda_for = read_pyproject_toml()
        if setup:
            self.install_any("pyproject.toml#build-system", toml_setup_reqs,
                             use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

        # Read test requirements from setup.cfg
        setup_cfg = read_setuptools_cfg()
        if setup:
            self.install_any("setup.cfg#setup_requires", setup_cfg.setup_requires,
                             use_conda_for=toml_use_conda_for, versions_dct=versions_dct)
        if install:
            self.install_any("setup.cfg#install_requires", setup_cfg.install_requires,
                             use_conda_for=toml_use_conda_for, versions_dct=versions_dct)
        if tests:
            self.install_any("setup.cfg#tests_requires", setup_cfg.tests_requires,
                             use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

        if phase is not None:
            self.install_any(phase, phase_reqs, use_conda_for=toml_use_conda_for, versions_dct=versions_dct)

    def install_any(self,
                    phase_name: str,
                    pkgs: Sequence[str],
                    use_conda_for: Sequence[str] = (),
                    versions_dct: Dict[str, str] = None,
                    logfile: Union[bool, str, Path] = True,
                    ):
        """Install the `pkgs` provided with `session.install(*pkgs)`, except for those present in `use_conda_for`"""

        nox_logger.debug("\nAbout to install *%s* requirements: %s.\n "
                         "Conda pkgs are %s" % (phase_name, pkgs, use_conda_for))

        # use the provided versions dictionary to update the versions
        if versions_dct is None:
            versions_dct = dict()
        pkgs = [pkg + versions_dct.get(pkg, "") for pkg in pkgs if versions_dct.get(pkg, "") != DONT_INSTALL]

        # install on conda... if the session uses conda backend
        if not isinstance(self.virtualenv, nox.virtualenv.CondaEnv):
            conda_pkgs = []
        else:
            conda_pkgs = [pkg_req for pkg_req in pkgs if any(get_req_pkg_name(pkg_req) == c for c in use_conda_for)]
            if len(conda_pkgs) > 0:
                nox_logger.info("[%s] Installing requirements with conda: %s" % (phase_name, conda_pkgs))
                self.conda_install2(*conda_pkgs, logfile=logfile)

        pip_pkgs = [pkg_req for pkg_req in pkgs if pkg_req not in conda_pkgs]
        # safety: make sure that nothing went modified or forgotten
        assert set(conda_pkgs).union(set(pip_pkgs)) == set(pkgs)
        if len(pip_pkgs) > 0:
            nox_logger.info("[%s] Installing requirements with pip: %s" % (phase_name, pip_pkgs))
            self.install2(*pip_pkgs, logfile=logfile)

    def conda_install2(self,
                       *conda_pkgs,
                       logfile: Union[bool, str, Path] = True,
                       **kwargs
                       ):
        """
        Same as session.conda_install() but with support for `logfile`.

        :param conda_pkgs:
        :param logfile: None/False (normal nox behaviour), or True (using nox file handler), or a file path.
        :return:
        """
        return self.conda_install(*conda_pkgs, logfile=logfile, **kwargs)

    def install2(self,
                 *pip_pkgs,
                 logfile: Union[bool, str, Path] = True,
                 **kwargs
                 ):
        """
        Same as session.install() but with support for `logfile`.

        :param pip_pkgs:
        :param logfile: None/False (normal nox behaviour), or True (using nox file handler), or a file path.
        :return:
        """
        return self.install(*pip_pkgs, logfile=logfile, **kwargs)

    def get_session_id(self):
        """Return the session id"""
        return Path(self.bin).name

    @classmethod
    def is_power_session(cls, session: Session):
        return PowerSession.install2.__name__ in session.__dict__

    @classmethod
    def patch(cls, session: Session):
        """
        Add all methods from this class to the provided object.
        Note that we could instead have created a proper proxy... but complex for not a lot of benefit.
        :param session:
        :return:
        """
        if not cls.is_power_session(session):
            for m_name, m in cls.__dict__.items():
                if not isfunction(m):
                    continue
                if m is cls.patch:
                    continue
                if not hasattr(session, m_name):
                    setattr(session.__class__, m_name, m)

        return True


# ------------- requirements related


def read_pyproject_toml():
    """
    Reads the `pyproject.toml` and returns

     - a list of setup requirements from [build-system] requires
     - sub-list of these requirements that should be installed with conda, from [tool.my_conda] conda_packages
    """
    if os.path.exists("pyproject.toml"):
        import toml
        nox_logger.debug("\nA `pyproject.toml` file exists. Loading it.")
        pyproject = toml.load("pyproject.toml")
        requires = pyproject['build-system']['requires']
        conda_pkgs = pyproject['tool']['conda']['conda_packages']
        return requires, conda_pkgs
    else:
        raise FileNotFoundError("No `pyproject.toml` file exists. No dependency will be installed ...")


SetupCfg = namedtuple('SetupCfg', ('setup_requires', 'install_requires', 'tests_requires'))


def read_setuptools_cfg():
    """
    Reads the `setup.cfg` file and extracts the various requirements lists
    """
    # see https://stackoverflow.com/a/30679041/7262247
    from setuptools import Distribution
    dist = Distribution()
    dist.parse_config_files()

    # standard requirements
    options_dct = dist.get_option_dict('options')
    setup_reqs = options_dct['setup_requires'][1].strip().splitlines()
    install_reqs = options_dct['install_requires'][1].strip().splitlines()
    tests_reqs = options_dct['tests_require'][1].strip().splitlines()

    return SetupCfg(setup_requires=setup_reqs,
                    install_requires=install_reqs,
                    tests_requires=tests_reqs)


def get_req_pkg_name(r):
    """Return the package name part of a python package requirement.

    For example
    "funcsigs;python<'3.5'" will return "funcsigs"
    "pytest>=3" will return "pytest"
    """
    return r.replace('<', '=').replace('>', '=').replace(';', '=').split("=")[0]


# ------------- log related


def with_logfile(logs_dir: Path,
                 logfile_arg: str = "logfile",
                 logfile_handler_arg: str = "logfilehandler"
                 ):
    """ A decorator to inject a logfile"""

    def _decorator(f):
        # check the signature of f
        foo_sig = signature(f)
        needs_logfile_injection = logfile_arg in foo_sig.parameters
        needs_logfilehandler_injection = logfile_handler_arg in foo_sig.parameters

        # modify the exposed signature if needed
        new_sig = None
        if needs_logfile_injection:
            new_sig = remove_signature_parameters(foo_sig, logfile_arg)
        if needs_logfilehandler_injection:
            new_sig = remove_signature_parameters(foo_sig, logfile_handler_arg)

        @wraps(f, new_sig=new_sig)
        def _f_wrapper(**kwargs):
            # find the session arg
            session = kwargs['session']  # type: Session

            # add file handler to logger
            logfile = logs_dir / ("%s.log" % PowerSession.get_session_id(session))
            error_logfile = logfile.with_name("ERROR_%s" % logfile.name)
            success_logfile = logfile.with_name("SUCCESS_%s" % logfile.name)
            # delete old files if present
            for _f in (logfile, error_logfile, success_logfile):
                if _f.exists():
                    _f.unlink()

            # add a FileHandler to the logger
            logfile_handler = log_to_file(logfile)

            # inject the log file / log file handler in the args:
            if needs_logfile_injection:
                kwargs[logfile_arg] = logfile
            if needs_logfilehandler_injection:
                kwargs[logfile_handler_arg] = logfile_handler

            # finally execute the session
            try:
                res = f(**kwargs)
            except Exception as e:
                # close and detach the file logger and rename as ERROR_....log
                remove_file_logger()
                logfile.rename(error_logfile)
                raise e
            else:
                # close and detach the file logger and rename as SUCCESS_....log
                remove_file_logger()
                logfile.rename(success_logfile)
                return res

        return _f_wrapper

    return _decorator


def log_to_file(file_path: Union[str, Path]
                ):
    """
    Closes and removes all file handlers from the nox logger,
    and add a new one to the provided file path

    :param file_path:
    :return:
    """
    for h in list(nox_logger.handlers):
        if isinstance(h, logging.FileHandler):
            h.close()
            nox_logger.removeHandler(h)
    fh = logging.FileHandler(str(file_path), mode='w')
    nox_logger.addHandler(fh)
    return fh


def get_current_logfile_handler():
    """
    Returns the current unique log file handler (see `log_to_file`)
    """
    for h in list(nox_logger.handlers):
        if isinstance(h, logging.FileHandler):
            return h
    return None


def get_log_file_stream():
    """
    Returns the output stream for the current log file handler if any (see `log_to_file`)
    """
    h = get_current_logfile_handler()
    if h is not None:
        return h.stream
    return None


def remove_file_logger():
    """
    Closes and detaches the current logfile handler
    :return:
    """
    h = get_current_logfile_handler()
    if h is not None:
        h.close()
        nox_logger.removeHandler(h)


# ------------ environment grid / parametrization related


def inject_envs_params(envs):
    param_names = None
    for env_py, env_params in envs.items():
        if param_names is None:
            param_names = set(env_params.keys())
        else:
            if param_names != set(env_params.keys()):
                raise ValueError("Environment %r parameters %r does not match parameters in the first environment: %r"
                                 % (env_py, param_names, set(env_params.keys())))

    def _decorator(f):
        # check the signature of f
        foo_sig = signature(f)
        missing = param_names - set(foo_sig.parameters)
        if len(missing) > 0:
            raise ValueError("Session function %r does not contain environment parameter(s) %r" % (f.__name__, missing))

        # modify the exposed signature if needed
        new_sig = None
        if len(param_names) > 0:
            new_sig = remove_signature_parameters(foo_sig, *param_names)

        @wraps(f, new_sig=new_sig)
        def _f_wrapper(**kwargs):
            # find the session arg
            session = kwargs['session']    # type: Session

            # get the versions to use for this environment
            try:
                params_dct = envs[session.python]
            except KeyError:
                nox_logger.warning(
                    "Skipping configuration, this is not supported in python version %r" % session.python)
                return

            # inject the parameters in the args:
            kwargs.update(params_dct)

            # finally execute the session
            return f(**kwargs)

        return _f_wrapper

    return _decorator


# ----------- other goodies


def rm_file(folder: Union[str, Path]
            ):
    """Since on windows Path.unlink throws permission error sometimes, os.remove is preferred."""
    if isinstance(folder, str):
        folder = Path(folder)

    if folder.exists():
        os.remove(str(folder))
        # Folders.site.unlink()  --> possible PermissionError


def rm_folder(folder: Union[str, Path]
              ):
    """Since on windows Path.unlink throws permission error sometimes, shutil is preferred."""
    if isinstance(folder, str):
        folder = Path(folder)

    if folder.exists():
        shutil.rmtree(str(folder))
        # Folders.site.unlink()  --> possible PermissionError


# --- the patch of popen able to tee to logfile --


import nox.popen as nox_popen_module
orig_nox_popen = nox_popen_module.popen


class LogFileStreamCtx:
    def __init__(self, logfile_stream):
        self.logfile_stream = logfile_stream

    def __enter__(self):
        return self.logfile_stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def patched_popen(
    args: Sequence[str],
    env: Mapping[str, str] = None,
    silent: bool = False,
    stdout: Union[int, IO] = None,
    stderr: Union[int, IO] = subprocess.STDOUT,
    logfile: Union[bool, str, Path] = None,
    **kwargs
) -> Tuple[int, str]:
    """
    Our patch of nox.popen.popen().

    Current behaviour in `nox` is

     - when `silent=True` (default), process err is redirected to STDOUT and process out is captured in a PIPE and sent
       to the logger (that does not displaying it :) )

     - when `silent=False` (explicitly set, or when nox is run with verbose flag), process out and process err are both
       redirected to STDOUT.

    Our implementation allows us to be a little more flexible:

     - if logfile is True or a string/Path, both process err and process out are both TEE-ed to logfile
     - at the same time, the above behaviour remains.

    :param args:
    :param env:
    :param silent:
    :param stdout:
    :param stderr:
    :param logfile: None/False (normal nox behaviour), or True (using nox file handler), or a file path.
    :return:
    """
    logfile_stream = get_log_file_stream()

    if logfile in (None, False) or (logfile is True and logfile_stream is None):
        # execute popen as usual
        return orig_nox_popen(args=args, env=env, silent=silent, stdout=stdout, stderr=stderr, **kwargs)

    else:
        # we'll need to tee the popen
        if logfile is True:
            ctx = LogFileStreamCtx
        else:
            ctx = lambda _: open(logfile, "a")

        with ctx(logfile_stream) as log_file_stream:
            if silent and stdout is not None:
                raise ValueError(
                    "Can not specify silent and stdout; passing a custom stdout always silences the commands output in "
                    "Nox's log."
                )

            shell = kwargs.get("shell", False)
            if shell:
                raise ValueError("Using shell=True is not yet supported with async streaming to log files")

            if stdout is not None or stderr is not subprocess.STDOUT:
                raise ValueError("Using custom streams is not yet supported with async popen")

            # old way
            # proc = subprocess.Popen(args, env=env, stdout=stdout, stderr=stderr)

            # New way: use asyncio to stream correctly
            # Note: if keyboard interrupts do not work we should check
            #  https://mail.python.org/pipermail/async-sig/2017-August/000374.html maybe or the following threads.

            # define the async coroutines
            async def async_popen():
                process = await asyncio.create_subprocess_exec(*args, env=env, stdout=asyncio.subprocess.PIPE,
                                                               stderr=asyncio.subprocess.PIPE, **kwargs)

                # bind the out and err streams - see https://stackoverflow.com/a/59041913/7262247
                # to mimic nox behaviour we only use a single capturing list
                outlines = []
                await asyncio.wait([
                    # process out is only redirected to STDOUT if not silent
                    _read_stream(process.stdout, lambda l: tee(l, sinklist=outlines, sinkstream=log_file_stream,
                                                               quiet=silent, verbosepipe=sys.stdout)),
                    # process err is always redirected to STDOUT (quiet=False) with a specific label
                    _read_stream(process.stderr, lambda l: tee(l, sinklist=outlines, sinkstream=log_file_stream,
                                                               quiet=False, verbosepipe=sys.stdout, label="ERR:"))
                ])
                return_code = await process.wait()  # make sur the process has ended and retrieve its return code
                return return_code, outlines

            # run the coroutine in the event loop
            loop = asyncio.get_event_loop()
            return_code, outlines = loop.run_until_complete(async_popen())

            # just in case, flush everything
            log_file_stream.flush()
            sys.stdout.flush()
            sys.stderr.flush()

            if silent:
                # same behaviour as in nox: this will be passed to the logger, and it will act depending on verbose flag
                out = "\n".join(outlines) if len(outlines) > 0 else ""
            else:
                # already written to stdout, no need to capture
                out = ""

            return return_code, out


async def _read_stream(stream, callback):
    """Helper async coroutine to read from a stream line by line and write them in callback"""
    while True:
        line = await stream.readline()
        if line:
            callback(line)
        else:
            break


def tee(linebytes, sinklist, sinkstream, verbosepipe, quiet, label=""):
    """
    Helper routine to read a line, decode it, and append it to several sinks:

     - an optional `sinklist` list that will receive the decoded string in its "append" method
     - an optional `sinkstream` stream that will receive the decoded string in its "writelines" method
     - an optional `verbosepipe` stream that will receive only when quiet=False, the decoded string through a print

    append it to the sink, and if quiet=False, write it to pipe too.
    """
    line = linebytes.decode('utf-8').rstrip()

    if sinklist is not None:
        sinklist.append(line)

    if sinkstream is not None:
        sinkstream.write(line + "\n")
        sinkstream.flush()

    if not quiet and verbosepipe is not None:
        print(label, line, file=verbosepipe)
        verbosepipe.flush()


def patch_popen():
    nox_popen_module.popen = patched_popen

    from nox.command import popen
    if popen is not patched_popen:
        nox.command.popen = patched_popen

    # change event loop on windows
    # see https://stackoverflow.com/a/44639711/7262247
    # and https://docs.python.org/3/library/asyncio-platforms.html#subprocess-support-on-windows
    if 'win32' in sys.platform:
        # Windows specific event-loop policy & cmd
        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
        # cmds = [['C:/Windows/system32/HOSTNAME.EXE']]

    # loop = asyncio.ProactorEventLoop()
    # asyncio.set_event_loop(loop)


patch_popen()



import warnings
from ast import literal_eval
from getpass import getpass
import io
import os
from shutil import copyfile
from threading import Lock

try:
    # Python 3
    from urllib.parse import urlparse, parse_qs
except ImportError:
    # Python 2
    from urlparse import urlparse, parse_qs

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path  # python 2

import sys

if sys.version_info < (3,):
    # py2: we need a version of open that supports encoding and nline endings
    # See https://stackoverflow.com/a/10975371/7262247
    from io import open
    from StringIO import StringIO
    string_types = (str, bytes)
else:
    from io import StringIO
    string_types = (str,)

try:
    from json import loads, JSONDecodeError, dumps
except ImportError:
    # python 2
    from json import loads, dumps
    JSONDecodeError = ValueError
try:
    FileNotFoundError
except NameError:
    # python 2
    FileNotFoundError = IOError

from requests import Session, HTTPError

try:
    # python 3
    # noinspection PyCompatibility
    from urllib.parse import quote
except ImportError:
    # python 2
    # noinspection PyUnresolvedReferences
    from urllib import quote

try:
    # noinspection PyUnresolvedReferences
    from typing import Dict, Union, Iterable
except ImportError:
    pass

CACHE_ROOT_FOLDER = ".odsclient"
CACHE_ENCODING = "utf-8"
ODS_BASE_URL_TEMPLATE = "https://%s.opendatasoft.com"
ENV_ODS_APIKEY = 'ODS_APIKEY'
KR_DEFAULT_USERNAME = 'apikey_user'


class ODSClient(object):
    """
    An `ODSClient` is a client for a given OpenDataSoft (ODS) platform. By default the target platform base url is
    `https://<platform_id>.opendatasoft.com` with `platform_id='public'`. One can change either customize the platform
    id through the `platform_id` constructor argument, or the whole base url with `base_url`.

    A client instance offers methods to interact with the various ODS API. Currently three high-level methods are
    provided: `<client>.get_whole_dataset(dataset_id, ...)`, `<client>.get_whole_dataframe(dataset_id, ...)`
    and `<client>.push_dataset_realtime(dataset_id, ...)`.

    A file cache can be activated for the two `get` methods by setting `file_cache` to `True` or to a path-like (string
    or `Path`) indicating a custom cache root path. `True` will use the default cache root folder `.odsclient`.
    `<client>.get_cached_dataset_entry` can be used to get a `CacheEntry` object representing the (possibly
    non-existing) cache entry for a given dataset.

    You can customize the `requests.Session` object used for the HTTPS transport using `requests_session`. If you do so,
    remember to close it yourself or to switch `auto_close_session` to True.

    A client is meant to use a single api key at a time, or none. You can force the api key to be mandatory using
    `enforce_apikey=True`. There are 4 ways to pass an api key, they are used in the following order:

     - explicitly with the `apikey` argument

     - through a text file containing the key. This file if present should be named `ods.apikey` (name can be changed
       using `apikey_filepath`, it does not make the file mandatory)

     - if `keyring` is installed (`pip install keyring`), an apikey can be created as an entry in it for service
       `<base_url>` and username `'apikey_user'`. `keyring` leverages your OS' vault (Windows Credential Locker,
       macOS Keychain, etc. This is the **most secure** method available. You can override the default keyring entry
       username with the `keyring_entries_username=...` argument. You can easily add or remove an entry in the keyring
       through the OS interface, with the `odskeys` commandline utility (`odskeys --help`) or with the
       `<client>.store_apikey_in_keyring` / `<client>.get_apikey_from_keyring` / `<client>.remove_apikey_from_keyring`
       methods.

     - through the `'ODS_APIKEY'` OS environment variable. It should either contain the key without quotes or a
       dict-like structure where keys can either be `platform_id`, `base_url`, or the special fallback key `'default'`

    For debugging purposes, you may wish to use `<client>.get_apikey()` to check if the api key that is actually used
    is the one you think you have configured through one of the above methods.

    """

    def __del__(self):
        """
        This is called when the garbage collector deletes this object.
        Let's use this opportunity to close the requests Session to avoid
        leaving hanging Sockets, see https://github.com/smarie/python-odsclient/issues/27
        """
        if self.auto_close_session and self.session is not None:
            try:
                # close the underlying `requests.Session`
                self.session.close()
            except Exception as e:
                warnings.warn("Error while closing session: %r" % e)

    def __init__(self,
                 platform_id='public',                          # type: str
                 base_url=None,                                 # type: str
                 enforce_apikey=False,                          # type: bool
                 apikey=None,                                   # type: str
                 apikey_filepath='ods.apikey',                  # type: Union[str, Path]
                 use_keyring=True,                              # type: bool
                 keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                 requests_session=None,                         # type: Session
                 auto_close_session=None                        # type: bool
                 ):
        """
        Constructor for `ODSClient`s

        :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
            https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
            https://public.opendatasoft.com
        :param base_url: an explicit base url to use instead of the one generated from `platform_id`
        :param enforce_apikey: an optional boolean indicating if an error should be raised if no apikey is found at all
            (not in the explicit argument, not in a file, environment variable, nor keyring) (default `False`)
        :param apikey: an explicit api key as a string.
        :param apikey_filepath: the path that should be used to look for api keys on the file system. Such files are
            optional, other (safer) methods exist to pass the api key, see documentation for details.
        :param use_keyring: an optional boolean (default `True`) specifying whether the `keyring` library should be
            used to lookup existing api keys. Keys should be stored using `store_apikey_in_keyring()`.
        :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
            the base url for the service id, however the user name can be anything. By default we use a string:
            'apikey_user'.
        :param requests_session: an optional `Session` object to use (from `requests` lib). If `None` is provided,
            a new `Session` will be used and deleted when this object is garbaged out. If a custom object is provided,
            you should close it yourself or switch `auto_close_session` to `True` explicitly.
        :param auto_close_session: an optional boolean indicating if `self.session` should be closed when this object
            is garbaged out. By default this is `None` and means "`True` if no custom `requests_session` is passed, else
            `False`"). Turning this to `False` can leave hanging Sockets unclosed.
        """
        # keyring option
        self.use_keyring = use_keyring
        self.keyring_entries_username = keyring_entries_username

        # Construct the base url
        if base_url is not None:
            if platform_id != 'public' and platform_id is not None:
                raise ValueError("Only one of `platform_id` and `base_url` should be provided. Received "
                                 "platform_id='%s' and base_url='%s'" % (platform_id, base_url))
            # remove trailing slashes
            while base_url.endswith('/'):
                base_url = base_url[:-1]
            self.base_url = base_url
            self.platform_id = None
        else:
            self.platform_id = platform_id
            self.base_url = ODS_BASE_URL_TEMPLATE % platform_id

        # Load apikey from file and validate it
        self.apikey_filepath = apikey_filepath
        if apikey is not None:
            # api key passed as argument
            if apikey_filepath != 'ods.apikey':
                raise ValueError("Only one of `apikey` and custom `apikey_filepath` should be provided.")
            self.apikey = apikey

        elif apikey_filepath is not None:
            try:
                # read the api key from the file
                with open(str(apikey_filepath)) as f:
                    self.apikey = f.read()
            except FileNotFoundError:
                self.apikey = None
            else:
                # remove trailing new lines or blanks if any
                self.apikey = self.apikey.rstrip()
        else:
            # no explicit api key. Environment variable will apply
            self.apikey = None

        if self.apikey is not None and len(self.apikey) == 0:
            raise ValueError('The provided api key is empty!')

        # checker flag
        self.enforce_apikey = enforce_apikey

        # create and store a session
        self.session = requests_session or Session()
        # auto-close behaviour
        if auto_close_session is None:
            # default: only auto-close if this session was created by us.
            auto_close_session = requests_session is None
        self.auto_close_session = auto_close_session

    def get_whole_dataframe(self,
                            dataset_id,                  # type: str
                            use_labels_for_header=True,  # type: bool
                            tqdm=False,                  # type: bool
                            block_size=1024,             # type: int
                            file_cache=False,            # type: bool
                            **other_opts
                            ):
        """
        Returns a dataset as a pandas dataframe. pandas must be installed.

        :param dataset_id:
        :param use_labels_for_header:
        :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
        :param block_size: an int block size used in streaming mode when tqdm is used
        :param file_cache: a boolean (default False) indicating whether the file should be written to a local cache
            `.odsclient/<base_url>_<dataset_id>.<format>`. Or a path-like object with the custom cache root folder.
        :param other_opts:
        :return:
        """
        try:
            import pandas as pd
        except ImportError as e:
            raise Exception("`get_whole_dataframe` requires `pandas` to be installed. [%s] %s" % (e.__class__, e))

        # Combine all the options
        opts = other_opts
        apikey = self.get_apikey()
        if apikey is not None:
            opts['apikey'] = apikey
        if use_labels_for_header is not None:
            opts['use_labels_for_header'] = use_labels_for_header

        # hardcoded
        if 'timezone' in opts:
            raise ValueError("'timezone' should not be specified with this method")
        if 'format' in opts:
            raise ValueError("'format' should not be specified with this method")
        opts['format'] = format = 'csv'
        if 'csv_separator' in opts:
            raise ValueError("'csv_separator' should not be specified with this method")
        opts['csv_separator'] = ';'

        # Cache usage
        if file_cache:
            # it can be a boolean or a path
            if file_cache is True:
                file_cache = CACHE_ROOT_FOLDER
            cached_file = self.get_cached_dataset_entry(dataset_id=dataset_id, format=format, cache_root=file_cache)
            try:
                # try to read the cached file in a thread-safe operation
                with cached_file.rw_lock:
                    cached_file.assert_exists()
                    df = pd.read_csv(str(cached_file.file_path), sep=';')
                    return df
            except CacheFileNotFoundError:
                pass  # does not exist. continue to query
        else:
            cached_file = None
        del file_cache

        # The URL to call
        url = self.get_download_url(dataset_id)

        # Execute call in stream mode with automatic content-type decoding
        result = self._http_call(url, params=opts, stream=True, decode=True)
        # print(iterable_to_stream(result.iter_content()).read())
        # noinspection PyTypeChecker

        if tqdm:
            from tqdm import tqdm as _tqdm

            total_size = int(result.headers.get('Content-Length', 0))
            with _tqdm(desc=url, total=total_size,
                       unit='B' if block_size == 1024 else 'it',
                       unit_scale=True,
                       unit_divisor=block_size
                       ) as bar:
                if not cached_file:
                    # Directly stream to memory with updates of the progress bar
                    df = pd.read_csv(iterable_to_stream(result.iter_content(), buffer_size=block_size, progressbar=bar),
                                     sep=';')
                else:
                    # stream to cache file and read the dataframe from the cache (use the lock to make sure it is here)
                    with cached_file.rw_lock:
                        cached_file.fill_from_iterable(result.iter_content(block_size), it_encoding=result.encoding,
                                                       progress_bar=bar, lock=False)
                        df = pd.read_csv(str(cached_file.file_path), sep=';')
        else:
            if not cached_file:
                # directly stream to memory dataframe
                df = pd.read_csv(iterable_to_stream(result.iter_content(block_size)), sep=';')
            else:
                # stream to cache file and read the dataframe from the cache (use the lock to make sure it is here)
                with cached_file.rw_lock:
                    cached_file.fill_from_iterable(result.iter_content(block_size), it_encoding=result.encoding,
                                                   lock=False)
                    df = pd.read_csv(str(cached_file.file_path), sep=';')

        return df

    # noinspection PyShadowingBuiltins
    def get_whole_dataset(self,
                          dataset_id,                  # type: str
                          format='csv',                # type: str
                          timezone=None,               # type: str
                          use_labels_for_header=True,  # type: bool
                          csv_separator=';',           # type: str
                          tqdm=False,                  # type: bool
                          to_path=None,                # type: Union[str, Path]
                          file_cache=False,            # type: bool
                          block_size=1024,             # type: int
                          **other_opts
                          ):
        """
        Returns a dataset as a csv string.

        :param dataset_id:
        :param format:
        :param timezone:
        :param use_labels_for_header:
        :param csv_separator: ';', ','...
        :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
        :param to_path: a string indicating the file path where to write the csv. In that case None is returned
        :param file_cache: a boolean (default False) indicating whether the file should be written to a local cache
            `.odsclient/<base_url>_<dataset_id>.<format>`. Or a path-like object with the custom cache root folder.
        :param block_size: an int block size used in streaming mode when to_csv or tqdm is used
        :param other_opts:
        :return:
        """

        # ------- To uncomment one day if headers and/or body are needed
        # headers = {'Authorization': ('Bearer ' + api_key)}
        #
        # if not (requestJsonBodyStr is None):
        #     # first encode the string as bytes using the charset
        #     charset = 'utf-8'
        #     json_body_encoded_with_charset = str.encode(requestJsonBodyStr, encoding=charset)
        #     headers['Content-Type'] = 'application/json; charset=' + charset
        # else:
        #     json_body_encoded_with_charset = None
        # ------------------

        # Combine all the options
        opts = other_opts
        apikey = self.get_apikey()
        if apikey is not None:
            opts['apikey'] = apikey
        if format is not None:
            opts['format'] = format
        if timezone is not None:
            opts['timezone'] = timezone
        if use_labels_for_header is not None:
            opts['use_labels_for_header'] = use_labels_for_header
        if csv_separator is not None:
            opts['csv_separator'] = csv_separator

        # The URL to call
        url = self.get_download_url(dataset_id)

        # Should we write anything to disk ?
        # -- Because it is the target
        if to_path is not None:
            if isinstance(to_path, string_types):
                to_path = Path(to_path)
            to_path.parent.mkdir(parents=True, exist_ok=True)  # make sure the parents exist

        # -- Because the cache is used
        if file_cache:
            # it can be a boolean or a path
            if file_cache is True:
                file_cache = CACHE_ROOT_FOLDER
            cached_file = self.get_cached_dataset_entry(dataset_id=dataset_id, format=format, cache_root=file_cache)
            try:
                # Do NOT call cached_file.exists(): not thread-safe
                if to_path is None:
                    return cached_file.read()          # this is atomic: thread-safe
                else:
                    cached_file.copy_to_file(to_path)  # this is atomic: thread-safe
                    return None
            except CacheFileNotFoundError:
                # does not exist. continue to query
                pass
        else:
            cached_file = None
        del file_cache

        # Execute call, since no cache was used
        result = None
        if not tqdm:
            if to_path is None:
                # We need to return a csv string, so load everything in memory
                result, content_type = self._http_call(url, params=opts, stream=False, decode=True)

                if cached_file:  # cache it in local cache if needed
                    cached_file.fill_from_str(txt_initial_encoding=content_type, decoded_txt=result)
            else:
                # No need to return a csv string: stream directly to csv file (no decoding/encoding)
                r = self._http_call(url, params=opts, stream=True, decode=False)
                with open(str(to_path), mode='wb') as f:
                    for data in r.iter_content(block_size):
                        f.write(data)

                if cached_file:  # cache it in local cache if needed
                    cached_file.fill_from_file(file_path=to_path, file_encoding=r.encoding)
        else:
            # Progress bar is needed: we need streaming mode
            r = self._http_call(url, params=opts, stream=True, decode=False)
            total_size = int(r.headers.get('Content-Length', 0))

            from tqdm import tqdm as _tqdm
            with _tqdm(desc=url, total=total_size,
                       unit='B' if block_size == 1024 else 'it',
                       unit_scale=True,
                       unit_divisor=block_size
                       ) as bar:
                if to_path is None:
                    result = io.StringIO()                     # stream to a string in memory
                    for data in r.iter_content(block_size):    # block by block
                        bar.update(len(data))                  # - update progress bar
                        result.write(data.decode(r.encoding))  # - decode with proper encoding
                    result = result.getvalue()

                    if cached_file:                            # cache it in local cache if needed
                        cached_file.fill_from_str(txt_initial_encoding=r.encoding, decoded_txt=result)
                else:
                    with open(str(to_path), 'wb') as f:          # stream to csv file in binary mode
                        for data in r.iter_content(block_size):  # block by block
                            bar.update(len(data))                # - update progress bar
                            f.write(data)                        # - direct copy (no decoding/encoding)

                    if cached_file:                              # cache it in local cache if needed
                        cached_file.fill_from_file(file_path=to_path, file_encoding=r.encoding)

            if total_size != 0 and bar.n != total_size:
                raise ValueError("ERROR, something went wrong")

        return result

    # noinspection PyShadowingBuiltins
    def push_dataset_realtime(self,
                              dataset_id,         # type: str
                              dataset,            # type: Union[str, pandas.DataFrame]
                              push_key,           # type: str
                              format='csv',       # type: str
                              csv_separator=';',  # type: str
                              **other_opts
                              ):
        """
        Pushes a Dataset. This functions accepts either a Pandas Dataframe or a CSV string with header included.

        :param dataset_id:
        :param dataset: The dataset to push as a list of dicts, where the dict keys are the column names
        :param push_key: The Push Key provided by the API for pushing this dataset. Warning: This key is independent
                         from the API key. It can be acquired from the Realtime Push API URL section in ODS.
        :param format: The format of the dataset to be pushed. Can be `pandas` or `csv`.
        :param csv_separator: CSV separator character in case of a csv dataset input.
        :returns: HTTP Response status
        """

        if format == 'pandas':
            try:
                import pandas as pd
            except ImportError as e:
                raise Exception("`push_dataset_realtime` with the `pandas` format requires `pandas` to be installed. [%s] %s" % (e.__class__, e))
            # noinspection PyStatementEffect
            dataset  # type:pandas.DataFrame
            request_body = dataset.to_json(orient='records')
        elif format == 'csv':
            try:
                import csv
            except ImportError as e:
                raise Exception("`push_dataset_realtime` with the `csv` format requires `csv` to be installed. [%s] %s" % (e.__class__, e))
            # noinspection PyStatementEffect
            dataset  # type:str
            csv_reader = csv.DictReader(StringIO(dataset), delimiter=csv_separator)
            request_body = dumps([r for r in csv_reader])
        else:
            raise ValueError("Dataset format must be either `pandas` or `csv`")

        # Combine all the options
        opts = other_opts
        opts['pushkey'] = push_key

        # The URL to call
        url = self.get_realtime_push_url(dataset_id)

        # Execute call
        return self._http_call(url, method='post', body=request_body, params=opts, decode=False)

    def store_apikey_in_keyring(self,
                                apikey=None  # type: str
                                ):
        """
        Convenience method to store a password in the OS keyring using `keyring` lib.

        This method is a shortcut for `keyring.set_password(<base_url>, <keyring_entries_username>, <apikey>)`.

        :param apikey: an explicit apikey string. If not provided, `getpass()` will be used to prompt the user for the
            api key
        :return:
        """
        import keyring
        if apikey is None:
            apikey = getpass(prompt="Please enter your api key: ")

        if apikey is None or len(apikey) == 0:
            raise ValueError("Empty api key provided.")

        keyring.set_password(self.base_url, self.keyring_entries_username, apikey)

    def remove_apikey_from_keyring(self):
        """
        Convenience method to remove a previously stored password in the OS keyring using `keyring` lib.

        :return:
        """
        import keyring
        keyring.delete_password(self.base_url, self.keyring_entries_username)

    def get_apikey_from_keyring(self, ignore_import_errors=False):
        """
        Looks for a keyring entry containing the api key and returns it.
        If not found, returns `None`

        :param ignore_import_errors: when this is `True`, the method will return `None` if `keyring` is not installed
            instead of raising an `ImportError`.
        :return:
        """
        if ignore_import_errors:
            try:
                import keyring
            except ImportError as e:
                # not installed: simply warn instead of raising exception
                warnings.warn("`keyring` is not installed but the `ODSClient` is configured to use it. You can either"
                              "set `use_keyring=False` explicitly or install keyring to disable this warning. "
                              "Caught: %r" % e)
                return None
        else:
            # do not catch any exception
            import keyring

        for _url in (self.base_url, self.base_url + '/'):
            apikey = keyring.get_password(_url, self.keyring_entries_username)
            if apikey is not None:
                return apikey

    def get_apikey_from_envvar(self):
        """
        Looks for the 'ODS_APIKEY' environment variable.

         - if it does not exist return None
         - otherwise if the env variable does not begin with '{', consider it as the key
         - if it begins with '{', loads it as a dict and find a match in it, in the following order:
           platform_id, base_url, 'default'

        If the found key is an empty string, a ValueError is raised.

        :return: the api key found in the 'ODS_APIKEY' env variable (possibly for this platform_id /
            base_url), or None if it does not exist.
        """
        try:
            env_api_key = os.environ[ENV_ODS_APIKEY]
        except KeyError:
            # no env var - return None
            return None

        if len(env_api_key) > 0 and env_api_key[0] == '{':
            # a dictionary: use ast.literal_eval: more permissive than json and as safe.
            apikeys_dct = literal_eval(env_api_key)
            if not isinstance(apikeys_dct, dict):
                raise TypeError("Environment variable contains something that is neither a str not a dict")

            # remove trailing slash in keys
            def _remove_trailing_slash(k):
                while k.endswith('/'):
                    k = k[:-1]
                return k

            apikeys_dct = {_remove_trailing_slash(k): v for k, v in apikeys_dct.items()}

            # Try to get a match in the dict: first platform id, then base url, then default
            if self.platform_id in apikeys_dct:
                env_api_key = apikeys_dct[self.platform_id]
            elif self.base_url in apikeys_dct:
                env_api_key = apikeys_dct[self.base_url]
            elif 'default' in apikeys_dct:
                env_api_key = apikeys_dct['default']
            else:
                return None

        if len(env_api_key) == 0:
            raise ValueError("Empty api key found in '%s' environment variable." % ENV_ODS_APIKEY)

        return env_api_key

    def get_apikey(self):
        """
        Returns the api key that this client currently uses.

        :return:
        """
        # 1- if there is an overridden api key, use it
        if self.apikey is not None:
            return self.apikey

        # 2- if keyring service is installed and contains an entry, use it
        if self.use_keyring:
            apikey = self.get_apikey_from_keyring(ignore_import_errors=True)
            if apikey is not None:
                return apikey

        # 3- check existence of the reference environment variable
        apikey = self.get_apikey_from_envvar()
        if apikey is not None:
            return apikey

        # 4- finally if no key was found, raise an exception if a key was required
        if self.enforce_apikey:
            raise NoODSAPIKeyFoundError(self)

    def get_download_url(self,
                         dataset_id  # type: str
                         ):
        # type: (...) -> str
        """

        :param dataset_id:
        :return:
        """
        return "%s/explore/dataset/%s/download/" % (self.base_url, dataset_id)

    def get_cached_dataset_entry(self,
                                 dataset_id,       # type: str
                                 format,           # type: str
                                 cache_root=None   # type: Union[str, Path]
                                 ):
        # type: (...) -> CacheEntry
        """
        Returns a `CacheEntry` for the given dataset
        :param dataset_id:
        :param format:
        :param cache_root:
        :return:
        """
        if self.platform_id is not None:
            p = self.platform_id
        else:
            p = baseurl_to_id_str(self.base_url)
        return CacheEntry(dataset_id=dataset_id, dataset_format=format, platform_pseudo_id=p, cache_root=cache_root)

    def get_realtime_push_url(self,
                              dataset_id,  # type: str
                              ):
        # type: (...) -> str
        """

        :param dataset_id:
        :return:
        """
        return "%s/api/push/1.0/%s/realtime/push/" % (self.base_url, dataset_id)

    def _http_call(self,
                   url,           # type: str
                   body=None,     # type: bytes
                   headers=None,  # type: Dict[str, str]
                   method='get',  # type: str
                   params=None,   # type: Dict[str, str]
                   decode=True,   # type: bool
                   stream=False   # type: bool
                   ):
        """
        Sub-routine for HTTP web service call. If Body is None, a GET is performed

        :param url:
        :param body:
        :param headers:
        :param method:
        :param params:
        :param decode: a boolean (default True) indicating if the contents should be automatically decoded following
            the content-type encoding received in the HTTP response. If this is True and stream=False (default), the
            function returns a tuple (body, content type)
        :param stream:
        :return: either a tuple (text, encoding) (if stream=False and decode=True), or the response object
        """
        try:
            # Send the request (DO NOT encode the params, this is done automatically)
            response = self.session.request(method, url, headers=headers, data=body, params=params, stream=stream)

            # Success ? Read status code, raise an HTTPError if status is error
            # status = int(response.status_code)
            response.raise_for_status()

            # detect a "wrong 200 but true 401" (unauthorized)
            if 'html' in response.headers['Content-Type']:
                raise InsufficientRightsForODSResourceError(response.headers, response.text)

            if not stream:
                if decode:
                    # Contents (encoding is automatically used to read the body when calling response.text)
                    result = response.text
                    return result, response.encoding
                else:
                    return response
            else:
                response.raw.decode_content = decode
                return response

        except HTTPError as error:
            try:
                body = error.response.text
                # {
                #   "errorcode": 10002,
                #   "reset_time": "2017-10-17T00:00:00Z",
                #   "limit_time_unit": "day",
                #   "call_limit": 10000,
                #   "error": "Too many requests on the domain. Please contact the domain administrator."
                # }
                details = loads(body)
            except JSONDecodeError:
                # error parsing the json payload?
                pass
            else:
                raise ODSException(error.response.status_code, error.response.headers, **details)

            raise error


class NoODSAPIKeyFoundError(Exception):
    """
    Raised when no api key was found (no explicit api key provided, no api key file, no env variable entry, no keyring
    entry)
    """

    def __init__(self,
                 odsclient  # type: ODSClient
                 ):
        self.odsclient = odsclient

    def __str__(self):
        return "ODS API key file not found, while it is marked as mandatory for this call (`enforce_apikey=True`). " \
               "It should either be put in a text file at path '%s', or in the `ODS_APIKEY` OS environment variable, " \
               "or (recommended, most secure) in the local `keyring`. " \
               "See documentation for details: %s. Note that you can generate an API key on this web page: " \
               "%s/account/my-api-keys/." \
               % (self.odsclient.apikey_filepath, "https://smarie.github.io/python-odsclient/#c-declaring-an-api-key",
                  self.odsclient.base_url)


class InsufficientRightsForODSResourceError(Exception):
    """
    Raised when a HTTP 200 is received from ODS together with an HTML page as body. This happens when api key is
    missing or does not grant the appropriate rights for the required resource.
    """

    def __init__(self, headers, contents):
        self.headers = headers
        self.contents = contents

    def __str__(self):
        return "An ODS query returned a HTTP 200 (OK) but with a html content-type. This is probably an " \
               "authentication problem, please check your api key using `get_apikey()`. " \
               "Headers:\n%s\nResponse:\n%s\n" % (self.headers, self.contents)


class ODSException(Exception):
    """
    An error returned by the ODS API
    """

    def __init__(self, status_code, headers, **details):
        """

        :param status_code:
        :param headers:
        :param details:
        """
        super(ODSException, self).__init__()
        self.status_code = status_code
        self.headers = headers
        self.error_msg = details['error']
        self.details = details

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "Request failed (%s): %s\nDetails: %s\nHeaders: %s" % (self.status_code, self.error_msg,
                                                                      self.details, self.headers)


def create_session_for_fiddler():
    # type: (...) -> Session
    return create_session_for_proxy(http_proxyhost='localhost', http_proxyport=8888,
                                    use_http_for_https_proxy=True, ssl_verify=False)


def create_session_for_proxy(http_proxyhost,                  # type: str
                             http_proxyport,                  # type: int
                             https_proxyhost=None,            # type: str
                             https_proxyport=None,            # type: int
                             use_http_for_https_proxy=False,  # type: bool
                             ssl_verify=None
                             ):
    # type: (...) -> Session
    """
    Helper method to configure the `requests` package to use the proxy fo your choice and adapt the SSL certificate
    validation accordingly. Note that this is only if you do not with to use the default configuration (inherited
    from environment variables, so you can use https://smarie.github.io/develop-behind-proxy/switching/#envswitcher)

    :param http_proxyhost: mandatory proxy host for http
    :param http_proxyport: mandatory proxy port for http
    :param https_proxyhost: optional proxy host for https. If none is provided, http_proxyhost will be used
    :param https_proxyport: optional proxy port for https. If none is provided, http_proxyport will be used
    :param use_http_for_https_proxy: optional, if set to true the http protocol will be used to initiate communications
        with the proxy even for https calls (then calls will be done in https as usual).
    :param ssl_verify: optional ssl verification parameter. It may either be the path to an additional certificate
        to trust (recommended), or a boolean to enable (default)/disable (not recommended ! use only in debug mode !)
        certificate validation.
        See here for details : http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification
    :return: a requests.Session object that you may use with the rest of the library
    """
    # config and fallback
    https_proxyhost = https_proxyhost if https_proxyhost is not None else http_proxyhost
    https_proxyport = https_proxyport if https_proxyport is not None else http_proxyport
    https_proxy_protocol = 'http' if use_http_for_https_proxy else 'https'

    s = Session()
    s.proxies = {
        'http': 'http://%s:%s' % (http_proxyhost, http_proxyport),
        'https': '%s://%s:%s' % (https_proxy_protocol, https_proxyhost, https_proxyport),
    }
    if not (ssl_verify is None):
        s.verify = ssl_verify

    # IMPORTANT : otherwise the environment variables will always have precedence over user settings
    s.trust_env = False

    return s


def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE, progressbar=None):
    """
    Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
    input stream.

    The stream implements Python 3's newer I/O API (available in Python 2's io module).
    For efficiency, the stream is buffered.

    Source: https://stackoverflow.com/a/20260030/7262247
    """

    class IterStream(io.RawIOBase):
        def __init__(self):
            self.leftover = None

        def readable(self):
            return True

        def readinto(self, b):
            try:
                ln = len(b)  # We're supposed to return at most this much
                chunk = self.leftover or next(iterable)
                output, self.leftover = chunk[:ln], chunk[ln:]
                b[:len(output)] = output
                if progressbar:
                    progressbar.update(len(output))
                return len(output)
            except StopIteration:
                return 0  # indicate EOF

    return io.BufferedReader(IterStream(), buffer_size=buffer_size)


class CacheFileNotFoundError(FileNotFoundError):
    pass


class CacheEntry(object):
    """
    Represents a cache entry for a dataset, under `cache_root` (default CACHE_ROOT_FOLDER).
    It may not exist.

    Access to the file are thread-safe (atomic) to avoid collisions while the file is updated.
    """
    __slots__ = ('dataset_id', 'dataset_format', 'platform_pseudo_id', '_cache_root', 'rw_lock')

    def __init__(self,
                 dataset_id,          # type: str
                 dataset_format,      # type: str
                 platform_pseudo_id,  # type: str
                 cache_root=None      # type: Union[str, Path]
                 ):
        """Constructor from a dataset id and an optional root cache path"""
        self.dataset_id = dataset_id
        self.dataset_format = dataset_format
        self.platform_pseudo_id = platform_pseudo_id
        self.rw_lock = Lock()

        if cache_root is None:
            self._cache_root = None
        else:
            if isinstance(cache_root, string_types):
                cache_root = Path(cache_root)
            self._cache_root = cache_root

    def __repr__(self):
        return "CacheEntry(path='%s')" % self.file_path

    def exists(self):
        """Return True if there is a file for this cache entry, at self.file_path. Note that in multithread context
        this does not ensure that this will remain True at next python code step :) """
        return self.file_path.exists()

    @property
    def cache_root(self):
        # type: (...) -> Path
        """The root folder of the cache where this entry sits"""
        return self._cache_root if self._cache_root is not None else CACHE_ROOT_FOLDER

    @property
    def file_path(self):
        # type: (...) -> Path
        """The file where this entry sits (it may exist or not)"""
        return Path("%s/%s/%s.%s" % (self.cache_root, self.platform_pseudo_id, self.dataset_id, self.dataset_format))

    def assert_exists(self):
        """Raises an error if the file does not exist"""
        if not self.exists():
            raise CacheFileNotFoundError("Cached file entry can not be read as it does not exist: '%s'" % self.file_path)

    def read(self):
        # type: (...) -> str
        """
        Returns a string read from the cached file.
        Preserve line endings thanks to newline='' see See https://stackoverflow.com/a/50996542/7262247
        """
        with self.rw_lock:  # potentially wait for ongoing write/read to be completed, and prevent others to happen
            self.assert_exists()
            with self.file_path.open(mode="rt", newline='', encoding=CACHE_ENCODING) as f:
                result = f.read()
            return result

    def copy_to_file(self,
                     file_path  # type: Union[str, Path]
                     ):
        """
        Copy this cached file to to_path. Note that it will be encoded using CACHE_ENCODING
        but a warning was already issued at dataset retrieval time if original encoding was different
        """
        with self.rw_lock:  # potentially wait for ongoing write/read to be completed, and prevent others to happen
            self.assert_exists()
            copyfile(str(self.file_path), str(file_path))

    def delete(self):
        """
        Removes this cache entry with thread-safe protection
        """
        with self.rw_lock:
            if not self.exists():
                warnings.warn("Can not delete file entry: file does not exist: '%s'" % self.file_path)
            else:
                os.remove(str(self.file_path))

    def prepare_for_writing(self):
        """
        Makes all parent directories if needed
        Issues a warning if the file exists and is therefore overridden
        """
        if self.exists():
            warnings.warn("Cached file entry already exists and will be overridden: '%s'" % self.file_path)

        # make sure the parents exist
        self.file_path.parent.mkdir(parents=True, exist_ok=True)

    def fill_from_str(self,
                      txt_initial_encoding,  # type: str
                      decoded_txt,           # type: str
                      ):
        """Writes a text string (already decoded) to the cache, according to the cache's encoding

        If the original encoding is not equal to the cache encoding, a warning is issued.
        """
        with self.rw_lock:  # potentially wait for ongoing write/read to be completed, and prevent others to happen
            self.prepare_for_writing()
            # Our cache uses utf-8 for all files, in order not to have to remember encodings to read back
            if txt_initial_encoding != CACHE_ENCODING:
                self.warn_encoding(original_encoding=txt_initial_encoding, cache_encoding=CACHE_ENCODING)

            # copy with the correct encoding
            self.file_path.write_bytes(decoded_txt.encode(CACHE_ENCODING))

    def _fill_from_it_no_lock(self,
                              it,                 # type: Iterable
                              it_encoding,        # type: str
                              progress_bar=None,
                              ):
        """The no-lock version of fill from iterable"""

        self.prepare_for_writing()
        if it_encoding == CACHE_ENCODING:
            # no encoding change: direct copy
            with open(str(self.file_path), 'wb') as f:  # stream to csv file in binary mode
                for data in it:  # block by block
                    if progress_bar:
                        progress_bar.update(len(data))  # - update progress bar
                    f.write(data)  # - direct copy (no decoding/encoding)
        else:
            # Our cache uses utf-8 for all files, in order not to have to remember encodings to read back
            self.warn_encoding(original_encoding=it_encoding, cache_encoding=CACHE_ENCODING)

            # we will need transcoding. Fully stream to memory string and dump to cache and datframe after
            csv_str_io = io.StringIO()  # stream to a string in memory
            for data in it:  # block by block
                if progress_bar:
                    progress_bar.update(len(data))  # - update progress bar
                csv_str_io.write(data.decode(it_encoding))  # - decode with proper encoding
            csv_str = csv_str_io.getvalue()

            # store in cache with proper encoding
            self.file_path.write_bytes(csv_str.encode(CACHE_ENCODING))

    def fill_from_iterable(self,
                           it,                # type: Iterable
                           it_encoding,       # type: str
                           progress_bar=None,
                           lock=True
                           ):
        """
        Fill this cache entry from an iterable of bytes
        :param it:
        :param it_encoding:
        :param progress_bar:
        :return: csv_str if it was streamed to memory in the process (if transcoding was needed)
        """
        if lock:
            with self.rw_lock:  # potentially wait for ongoing write/read to be completed, and prevent others to happen
                self._fill_from_it_no_lock(it=it, it_encoding=it_encoding, progress_bar=progress_bar)
        else:
            self._fill_from_it_no_lock(it=it, it_encoding=it_encoding, progress_bar=progress_bar)

    def fill_from_file(self,
                       file_path,      # type: Union[str, Path]
                       file_encoding,  # type: str
                       ):
        """Copies a file to the cache.

        If the original encoding is not equal to the cache encoding, conversion happens and a warning is issued.
        """
        with self.rw_lock:  # potentially wait for ongoing write/read to be completed, and prevent others to happen
            self.prepare_for_writing()
            if file_encoding == CACHE_ENCODING:
                # no encoding change: direct copy
                copyfile(str(file_path), str(self.file_path))
            else:
                # Our cache uses utf-8 for all files, in order not to have to remember encodings to read back
                self.warn_encoding(original_encoding=file_encoding, cache_encoding=CACHE_ENCODING)
                # read with newline-preserve and with original encoding
                with open(str(file_path), mode="rt", newline='', encoding=file_encoding) as f_src:
                    contents = f_src.read()
                # write with the cache encoding
                with open(str(self.file_path), mode='wt', encoding=CACHE_ENCODING) as f_dest:
                    f_dest.write(contents)

    def warn_encoding(self, original_encoding, cache_encoding):
        """
        Issues a warning when the original encoding was different from the one in the cache and a conversion occured
        """
        warnings.warn(
            "[odsclient-cache] Cached file for dataset %r will use %r encoding while original encoding on "
            " ODS was %r. This will most probably have no side effects except if your dataset"
            " contains characters that can not be encoded in utf-8 such as old/alternative"
            " forms of east asian kanji. See https://en.wikipedia.org/wiki/Unicode#Issues"
            % (self.dataset_id, cache_encoding, original_encoding))


def baseurl_to_id_str(base_url):
    """ Transform an ODS platform url into an identifier string usable for example as file/folder name"""

    o = urlparse(base_url)

    # start with host name
    result_str = o.netloc

    # simplify the public ODS site
    if result_str.endswith(".opendatasoft.com"):
        result_str = result_str.replace(".opendatasoft.com", "")

    # optionally add custom sub-path
    if o.path and o.path != "/":
        _path = o.path.replace("/", "_")

        # ensure trailing _
        if not _path.startswith("_"):
            _path = "_" + _path

        # ensure no ending _
        if _path.endswith("_"):
            _path = _path[:-1]

        result_str += _path

    return result_str



"""
To understand this project's build structure

 - This project uses setuptools, so it is declared as the build system in the pyproject.toml file
 - We use as much as possible `setup.cfg` to store the information so that it can be read by other tools such as `tox`
   and `nox`. So `setup.py` contains **almost nothing** (see below)
   This philosophy was found after trying all other possible combinations in other projects :)
   A reference project that was inspiring to make this move : https://github.com/Kinto/kinto/blob/master/setup.cfg

See also:
  https://setuptools.readthedocs.io/en/latest/setuptools.html#configuring-setup-using-setup-cfg-files
  https://packaging.python.org/en/latest/distributing.html
  https://github.com/pypa/sampleproject
"""
from setuptools import setup


# (1) check required versions (from https://medium.com/@daveshawley/safely-using-setup-cfg-for-metadata-1babbe54c108)
import pkg_resources

pkg_resources.require("setuptools>=39.2")
pkg_resources.require("setuptools_scm")


# (2) Generate download url using git version
from setuptools_scm import get_version  # noqa: E402

URL = "https://github.com/smarie/python-odsclient"
DOWNLOAD_URL = URL + "/tarball/" + get_version()


# (3) Call setup() with as little args as possible
setup(
    download_url=DOWNLOAD_URL,
    use_scm_version={
        "write_to": "odsclient/_version.py"
    },  # we can't put `use_scm_version` in setup.cfg yet unfortunately
)



import os
from glob import glob
from shutil import rmtree

try:
    # noinspection PyUnresolvedReferences
    from typing import Union
except ImportError:
    pass

try:
    from pathlib import Path
except ImportError:
    # do not care: only used for type hinting
    pass

from requests import Session

from odsclient.core import KR_DEFAULT_USERNAME, ODSClient, CACHE_ROOT_FOLDER, baseurl_to_id_str, CacheEntry


def store_apikey_in_keyring(platform_id='public',                          # type: str
                            base_url=None,                                 # type: str
                            keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                            apikey=None,                                   # type: str
                            ):
    """
    Convenience method to store a password in the OS keyring using `keyring` lib.
    It is equivalent to `ODSClient(...).store_apikey_in_keyring(apikey)`

    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param base_url: an explicit base url to use instead of the one generated from `platform_id`
    :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
        the base url for the service id, however the user name can be anything. By default we use a string:
        'apikey_user'.
    :param apikey: an explicit apikey string. If not provided, `getpass()` will be used to prompt the user for the
            api key
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, keyring_entries_username=keyring_entries_username)
    client.store_apikey_in_keyring(apikey=apikey)


def get_apikey_from_keyring(platform_id='public',                          # type: str
                            base_url=None,                                 # type: str
                            keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                            ):
    """
    Convenience method to get a previously stored password in the OS keyring using `keyring` lib.
    It is equivalent to `ODSClient(...).get_apikey_from_keyring()`

    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, keyring_entries_username=keyring_entries_username)
    return client.get_apikey_from_keyring(ignore_import_errors=False)


def remove_apikey_from_keyring(platform_id='public',                          # type: str
                               base_url=None,                                 # type: str
                               keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                               ):
    """
    Convenience method to remove a previously stored password in the OS keyring using `keyring` lib.
    It is equivalent to `ODSClient(...).remove_apikey_from_keyring()`

    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param base_url: an explicit base url to use instead of the one generated from `platform_id`
    :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
        the base url for the service id, however the user name can be anything. By default we use a string:
        'apikey_user'.
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, keyring_entries_username=keyring_entries_username)
    client.remove_apikey_from_keyring()


def get_apikey(platform_id='public',                          # type: str
               base_url=None,                                 # type: str
               apikey_filepath='ods.apikey',                  # type: str
               use_keyring=True,                              # type: bool
               keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
               ):
    # type: (...) -> str
    """
    Convenience method to check what is the api key used by ods clients.
    It is equivalent to `ODSClient(...).get_apikey()`

    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param base_url: an explicit base url to use instead of the one generated from `platform_id`
    :param apikey_filepath: the path that should be used to look for api keys on the file system. Such files are
        optional, other (safer) methods exist to pass the api key, see documentation for details.
    :param use_keyring: an optional boolean specifying whether the `keyring` library should be used to lookup
        existing api keys. Keys should be stored as `keyring.set_password(<base_url>, 'apikey', <apikey>)` where
        `<base_url>` should not contain any trailing slash.
    :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
        the base url for the service id, however the user name can be anything. By default we use a string:
        'apikey_user'.
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, apikey_filepath=apikey_filepath,
                       use_keyring=use_keyring, keyring_entries_username=keyring_entries_username)
    return client.get_apikey()


def get_whole_dataframe(dataset_id,                                    # type: str
                        use_labels_for_header=True,                    # type: bool
                        tqdm=False,                                    # type: bool
                        block_size=1024,                               # type: int
                        file_cache=False,                              # type: bool
                        platform_id='public',                          # type: str
                        base_url=None,                                 # type: str
                        enforce_apikey=False,                          # type: bool
                        apikey=None,                                   # type: str
                        apikey_filepath='ods.apikey',                  # type: Union[str, Path]
                        use_keyring=True,                              # type: bool
                        keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                        requests_session=None,                         # type: Session
                        auto_close_session=None,                       # type: bool
                        **other_opts
                        ):
    """
    Shortcut method for ODSClient(...).get_whole_dataframe(...)
    Returns a dataset as a pandas dataframe. pandas must be installed.

    :param dataset_id:
    :param use_labels_for_header:
    :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
    :param block_size: an int block size used in streaming mode when to_csv or tqdm is used
    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param base_url: an explicit base url to use instead of the one generated from `platform_id`
    :param enforce_apikey: an optional boolean indicating if an error should be raised if no apikey is found at all
        (not in the explicit argument, not in a file, environment variable, nor keyring) (default `False`)
    :param apikey: an explicit api key as a string.
    :param apikey_filepath: the path that should be used to look for api keys on the file system. Such files are
        optional, other (safer) methods exist to pass the api key, see documentation for details.
    :param use_keyring: an optional boolean  (default `True`) specifying whether the `keyring` library should be used
        to lookup existing api keys. Keys should be stored using `store_apikey_in_keyring()`.
    :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
        the base url for the service id, however the user name can be anything. By default we use a string:
        'apikey_user'.
    :param requests_session: an optional `Session` object to use (from `requests` lib). If `None` is provided,
            a new `Session` will be used and deleted when this object is garbaged out. If a custom object is provided,
            you should close it yourself or switch `auto_close_session` to `True` explicitly.
    :param auto_close_session: an optional boolean indicating if `self.session` should be closed when this object
        is garbaged out. By default this is `None` and means "`True` if no custom `requests_session` is passed, else
        `False`"). Turning this to `False` can leave hanging Sockets unclosed.
    :param other_opts:
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, enforce_apikey=enforce_apikey, apikey=apikey,
                       apikey_filepath=apikey_filepath, use_keyring=use_keyring, auto_close_session=auto_close_session,
                       keyring_entries_username=keyring_entries_username, requests_session=requests_session)
    return client.get_whole_dataframe(dataset_id=dataset_id, use_labels_for_header=use_labels_for_header,
                                      tqdm=tqdm, block_size=block_size, file_cache=file_cache, **other_opts)


def clean_cache(dataset_id=None,   # type: str
                # format='csv',    # type: str
                platform_id=None,  # type: str
                base_url=None,     # type: str
                cache_root=None    # type: Union[str, Path]
                ):
    """
    Cleans the file cache

    :param dataset_id:
    :param platform_id:
    :param base_url:
    :return:
    """
    if dataset_id is not None:
        # clean a specific dataset on a specific platform
        path_pattern = get_cached_dataset_entry(dataset_id, format="*", platform_id=platform_id, base_url=base_url,
                                                cache_root=cache_root)
        for cached_file in glob(str(path_pattern.file_path)):
            print("[odsclient] Removing cached dataset entry for %r: %r" % (dataset_id, cached_file))
            os.remove(cached_file)
    else:
        if cache_root is None:
            cache_root = CACHE_ROOT_FOLDER
        else:
            cache_root = str(cache_root)

        if platform_id is not None:
            p = platform_id
        elif base_url is not None:
            p = baseurl_to_id_str(base_url)
        else:
            p = None

        if p is None:
            # clean the whole cache
            print("[odsclient] Removing entire cache folder %r" % cache_root)
            rmtree(cache_root, ignore_errors=True)
        else:
            # clean an entire platform cache
            path_to_delete = "%s/%s/" % (cache_root, p)
            print("[odsclient] Removing cache for platform %r: folder %r" % (p, path_to_delete))
            rmtree(path_to_delete, ignore_errors=True)


def get_cached_dataset_entry(dataset_id,            # type: str
                             format='csv',          # type: str
                             platform_id='public',  # type: str
                             base_url=None,         # type: str
                             cache_root=None        # type: Union[str, Path]
                             ):
    # type: (...) -> CacheEntry
    """
    Shortcut method for ODSClient(...).get_cached_dataset_entry(...)

    :param dataset_id:
    :param format:
    :param platform_id:
    :param base_url:
    :param cache_root:
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url)
    return client.get_cached_dataset_entry(dataset_id=dataset_id, format=format, cache_root=cache_root)


# noinspection PyShadowingBuiltins
def get_whole_dataset(dataset_id,                                    # type: str
                      format='csv',                                  # type: str
                      timezone=None,                                 # type: str
                      use_labels_for_header=True,                    # type: bool
                      csv_separator=';',                             # type: str
                      tqdm=False,                                    # type: bool
                      to_path=None,                                  # type: Union[str, Path]
                      file_cache=False,                              # type: bool
                      block_size=1024,                               # type: int
                      platform_id='public',                          # type: str
                      base_url=None,                                 # type: str
                      enforce_apikey=False,                          # type: bool
                      apikey=None,                                   # type: str
                      apikey_filepath='ods.apikey',                  # type: Union[str, Path]
                      use_keyring=True,                              # type: bool
                      keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                      requests_session=None,                         # type: Session
                      auto_close_session=None,                       # type: bool
                      **other_opts
                      ):
    """
    Shortcut method for ODSClient(...).get_whole_dataset(...)

    :param dataset_id:
    :param format:
    :param timezone:
    :param use_labels_for_header:
    :param csv_separator:
    :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
    :param to_path: a string indicating the file path where to write the dataset (csv or other format). In that case
        nothing is returned
    :param file_cache: a boolean (default False) indicating whether the file should be written to a local cache
        `.odsclient/<pseudo_platform_id>_<dataset_id>.<format>`. See `get_cached_datasset_entry` for details.
    :param block_size: an int block size used in streaming mode when to_csv or tqdm is used
    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param base_url: an explicit base url to use instead of the one generated from `platform_id`
    :param enforce_apikey: an optional boolean indicating if an error should be raised if no apikey is found at all
        (not in the explicit argument, not in a file, environment variable, nor keyring) (default `False`)
    :param apikey: an explicit api key as a string.
    :param apikey_filepath: the path that should be used to look for api keys on the file system. Such files are
        optional, other (safer) methods exist to pass the api key, see documentation for details.
    :param use_keyring: an optional boolean  (default `True`) specifying whether the `keyring` library should be used
        to lookup existing api keys. Keys should be stored using `store_apikey_in_keyring()`.
    :param keyring_entries_username: keyring stores secrets with a key made of a service id and a username. We use
        the base url for the service id, however the user name can be anything. By default we use a string:
        'apikey_user'.
    :param requests_session: an optional `Session` object to use (from `requests` lib). If `None` is provided,
            a new `Session` will be used and deleted when this object is garbaged out. If a custom object is provided,
            you should close it yourself or switch `auto_close_session` to `True` explicitly.
    :param auto_close_session: an optional boolean indicating if `self.session` should be closed when this object
        is garbaged out. By default this is `None` and means "`True` if no custom `requests_session` is passed, else
        `False`"). Turning this to `False` can leave hanging Sockets unclosed.
    :param other_opts:
    :return:
    """
    client = ODSClient(platform_id=platform_id, base_url=base_url, enforce_apikey=enforce_apikey, apikey=apikey,
                       apikey_filepath=apikey_filepath, use_keyring=use_keyring, auto_close_session=auto_close_session,
                       keyring_entries_username=keyring_entries_username, requests_session=requests_session)
    return client.get_whole_dataset(dataset_id=dataset_id, format=format, file_cache=file_cache,
                                    timezone=timezone, use_labels_for_header=use_labels_for_header,
                                    csv_separator=csv_separator, tqdm=tqdm, to_path=to_path, block_size=block_size,
                                    **other_opts)


def push_dataset_realtime(platform_id,        # type: str
                          dataset_id,         # type: str
                          dataset,            # type: Union[str, pandas.DataFrame]
                          push_key,           # type: str
                          format='csv',       # type: str
                          csv_separator=';',  # type: str
                          **other_opts
                          ):
    """
    Pushes a Dataset. This functions accepts either a Pandas Dataframe or a CSV string with header included.

    :param platform_id: the ods platform id to use. This id is used to construct the base URL based on the pattern
        https://<platform_id>.opendatasoft.com. Default is `'public'` which leads to the base url
        https://public.opendatasoft.com
    :param dataset_id:
    :param dataset: The dataset to push as a list of dicts, where the dict keys are the column names
    :param push_key: The Push Key provided by the API for pushing this dataset. Warning: This key is independent
                     from the API key. It can be acquired from the Realtime Push API URL section in ODS.
    :param format: The format of the dataset to be pushed. Can be `pandas` or `csv`.
    :param csv_separator: CSV separator character in case of a csv dataset input.
    :returns: HTTP Response status
    """

    client = ODSClient(platform_id=platform_id)
    return client.push_dataset_realtime(dataset_id=dataset_id,
                                        dataset=dataset,
                                        push_key=push_key,
                                        format=format,
                                        csv_separator=csv_separator,
                                        **other_opts)



import logging

import nox  # noqa
from pathlib import Path  # noqa
import sys

# add parent folder to python path so that we can import noxfile_utils.py
# note that you need to "pip install -r noxfile-requiterements.txt" for this file to work.
sys.path.append(str(Path(__file__).parent / "ci_tools"))
from nox_utils import PY27, PY37, PY36, PY35, PY38, power_session, rm_folder, rm_file, PowerSession  # noqa


pkg_name = "odsclient"
gh_org = "smarie"
gh_repo = "python-odsclient"

ALL_PY_VERSIONS = [PY38, PY37, PY36, PY35, PY27]

ENVS = {
    PY27: {"coverage": False, "pkg_specs": {"pip": ">10"}},
    PY35: {"coverage": False, "pkg_specs": {"pip": ">10"}},
    PY36: {"coverage": False, "pkg_specs": {"pip": ">19"}},
    PY38: {"coverage": False, "pkg_specs": {"pip": ">19"}},
    # IMPORTANT: this should be last so that the folder docs/reports is not deleted afterwards
    PY37: {"coverage": True, "pkg_specs": {"pip": ">19"}},  # , "pytest-html": "1.9.0"
}


# set the default activated sessions, minimal for CI
nox.options.sessions = ["tests"]  # , "docs", "gh_pages"
nox.options.reuse_existing_virtualenvs = True  # this can be done using -r
# if platform.system() == "Windows":  >> always use this for better control
nox.options.default_venv_backend = "conda"
# os.environ["NO_COLOR"] = "True"  # nox.options.nocolor = True does not work
# nox.options.verbose = True

nox_logger = logging.getLogger("nox")
# nox_logger.setLevel(logging.INFO)  NO !!!! this prevents the "verbose" nox flag to work !


class Folders:
    root = Path(__file__).parent
    runlogs = root / Path(nox.options.envdir or ".nox") / "_runlogs"
    runlogs.mkdir(parents=True, exist_ok=True)
    dist = root / "dist"
    site = root / "site"
    site_reports = site / "reports"
    reports_root = root / "docs" / "reports"
    test_reports = reports_root / "junit"
    coverage_reports = reports_root / "coverage"
    coverage_xml = coverage_reports / "coverage.xml"


@power_session(envs=ENVS, logsdir=Folders.runlogs)
def tests(session: PowerSession, coverage, pkg_specs):
    """Run the test suite, including test reports generation and coverage reports. """

    # As soon as this runs, we delete the target site and coverage files to avoid reporting wrong coverage/etc.
    rm_folder(Folders.site)
    rm_folder(Folders.reports_root)
    # delete the .coverage files if any (they are not supposed to be any, but just in case)
    rm_file(Folders.root / ".coverage")
    rm_file(Folders.root / "coverage.xml")

    # Did we receive a flag through positional arguments ? (nox -s tests -- <flag>)
    install_keyrings_alt = False
    if len(session.posargs) == 1:
        assert session.posargs[0] == "keyrings.alt"
        install_keyrings_alt = True
    elif len(session.posargs) > 1:
        raise ValueError("Only a single positional argument is accepted, received: %r" % session.posargs)

    # uncomment and edit if you wish to uninstall something without deleting the whole env
    # session.run2("pip uninstall pytest-asyncio --yes")

    # install all requirements
    session.install_reqs(setup=True, install=True, tests=True, versions_dct=pkg_specs)

    # install extra req
    if install_keyrings_alt:
        session.install2("keyrings.alt")

    # install self so that it is recognized by pytest
    session.run2("pip install -e . --no-deps")

    # check that it can be imported even from a different folder
    session.run2(['python', '-c', '"import os; os.chdir(\'./docs/\'); import %s"' % pkg_name])

    # finally run all tests
    if not coverage:
        # simple: pytest only
        session.run2("python -m pytest -v %s/tests/" % pkg_name)
    else:
        # coverage + junit html reports + badge generation
        session.install_reqs(phase="coverage", phase_reqs=["coverage", "pytest-html", "requests", "xunitparser"],
                             versions_dct=pkg_specs)

        # --coverage + junit html reports
        session.run2("coverage run --source {pkg_name} "
                     "-m pytest --junitxml={dst}/junit.xml --html={dst}/report.html -v {pkg_name}/tests/"
                     "".format(pkg_name=pkg_name, dst=Folders.test_reports))
        # session.run2("coverage report")  # this shows in terminal + fails under XX%, same as --cov-report term --cov-fail-under=70  # noqa
        session.run2("coverage xml -o {covxml}".format(covxml=Folders.coverage_xml))
        session.run2("coverage html -d {dst}".format(dst=Folders.coverage_reports))
        # delete this intermediate file, it is not needed anymore
        rm_file(Folders.root / ".coverage")

        # --generates the badge for the test results and fail build if less than x% tests pass
        nox_logger.info("Generating badge for tests coverage")
        session.run2("python ci_tools/generate-junit-badge.py 100 %s" % Folders.test_reports)


@power_session(python=[PY37])
def docs(session: PowerSession):
    """Generates the doc and serves it on a local http server. Pass '-- build' to build statically instead."""

    session.install_reqs(phase="docs", phase_reqs=["mkdocs-material", "mkdocs", "pymdown-extensions", "pygments"])

    if session.posargs:
        # use posargs instead of "serve"
        session.run2("mkdocs -f ./docs/mkdocs.yml %s" % " ".join(session.posargs))
    else:
        session.run2("mkdocs serve -f ./docs/mkdocs.yml")


@power_session(python=[PY37])
def publish(session: PowerSession):
    """Deploy the docs+reports on github pages. Note: this rebuilds the docs"""

    session.install_reqs(phase="mkdocs", phase_reqs=["mkdocs-material", "mkdocs", "pymdown-extensions", "pygments"])

    # possibly rebuild the docs in a static way (mkdocs serve does not build locally)
    session.run2("mkdocs build -f ./docs/mkdocs.yml")

    # check that the doc has been generated with coverage
    if not Folders.site_reports.exists():
        raise ValueError("Test reports have not been built yet. Please run 'nox -s tests-3.7' first")

    # publish the docs
    session.run2("mkdocs gh-deploy -f ./docs/mkdocs.yml")

    # publish the coverage - now in github actions only
    # session.install_reqs(phase="codecov", phase_reqs=["codecov", "keyring"])
    # # keyring set https://app.codecov.io/gh/<org>/<repo> token
    # import keyring  # (note that this import is not from the session env but the main nox env)
    # codecov_token = keyring.get_password("https://app.codecov.io/gh/<org>/<repo>>", "token")
    # # note: do not use --root nor -f ! otherwise "There was an error processing coverage reports"
    # session.run2('codecov -t %s -f %s' % (codecov_token, Folders.coverage_xml))


@power_session(python=[PY37])
def release(session: PowerSession):
    """Create a release on github corresponding to the latest tag"""

    # Get current tag using setuptools_scm and make sure this is not a dirty/dev one
    from setuptools_scm import get_version  # (note that this import is not from the session env but the main nox env)
    from setuptools_scm.version import guess_next_dev_version
    version = []

    def my_scheme(version_):
        version.append(version_)
        return guess_next_dev_version(version_)
    current_tag = get_version(".", version_scheme=my_scheme)

    # create the package
    session.install_reqs(phase="setup.py#dist", phase_reqs=["setuptools_scm"])
    rm_folder(Folders.dist)
    session.run2("python setup.py sdist bdist_wheel")

    if version[0].dirty or not version[0].exact:
        raise ValueError("You need to execute this action on a clean tag version with no local changes.")

    # Did we receive a token through positional arguments ? (nox -s release -- <token>)
    if len(session.posargs) == 1:
        # Run from within github actions - no need to publish on pypi
        gh_token = session.posargs[0]
        publish_on_pypi = False

    elif len(session.posargs) == 0:
        # Run from local commandline - assume we want to manually publish on PyPi
        publish_on_pypi = True

        # keyring set https://docs.github.com/en/rest token
        import keyring  # (note that this import is not from the session env but the main nox env)
        gh_token = keyring.get_password("https://docs.github.com/en/rest", "token")
        assert len(gh_token) > 0

    else:
        raise ValueError("Only a single positional arg is allowed for now")

    # publish the package on PyPi
    if publish_on_pypi:
        # keyring set https://upload.pypi.org/legacy/ your-username
        # keyring set https://test.pypi.org/legacy/ your-username
        session.install_reqs(phase="PyPi", phase_reqs=["twine"])
        session.run2("twine upload dist/* -u smarie")  # -r testpypi

    # create the github release
    session.install_reqs(phase="release", phase_reqs=["click", "PyGithub"])
    session.run2("python ci_tools/github_release.py -s {gh_token} "
                 "--repo-slug {gh_org}/{gh_repo} -cf ./docs/changelog.md "
                 "-d https://{gh_org}.github.io/{gh_repo}/changelog/ {tag}"
                 "".format(gh_token=gh_token, gh_org=gh_org, gh_repo=gh_repo, tag=current_tag))


# if __name__ == '__main__':
#     # allow this file to be executable for easy debugging in any IDE
#     nox.run(globals())



import pytest
from click.testing import CliRunner

from odsclient.keyring_cmds import odskeys


@pytest.mark.parametrize('platform_id, base_url', [(None, None),
                                                   ('hello', None),
                                                   (None, 'http://blouh/')])
def test_odskey(platform_id, base_url):
    """Basic tests for the CLI """

    other_args = []
    if platform_id is not None:
        assert base_url is None
        other_args += ['-p', platform_id]
        url_used = "https://%s.opendatasoft.com" % platform_id
    elif base_url is not None:
        other_args += ['-b', base_url]
        url_used = base_url[0:-1] if base_url.endswith('/') else base_url
    else:
        url_used = "https://public.opendatasoft.com"
    msg = "platform url '%s'" % url_used

    runner = CliRunner()

    result = runner.invoke(odskeys, ['remove'] + other_args)
    assert result.exit_code == 0
    assert result.output == "No api key registered for %s\n" % msg

    result = runner.invoke(odskeys, ['get'] + other_args)
    assert result.exit_code == 0
    assert result.output == "No api key registered for %s\n" % msg

    result = runner.invoke(odskeys, ['set', '-k', 'blah'] + other_args)
    assert result.exit_code == 0
    assert result.output == "Api key defined successfully for %s\n" % msg

    result = runner.invoke(odskeys, ['get'] + other_args)
    assert result.exit_code == 0
    assert result.output == "Api key found for %s: blah\n" % msg

    result = runner.invoke(odskeys, ['remove'] + other_args)
    assert result.exit_code == 0
    assert result.output == "Api key removed successfully for %s\n" % msg

    result = runner.invoke(odskeys, ['get'] + other_args)
    assert result.exit_code == 0
    assert result.output == "No api key registered for %s\n" % msg
